#!/usr/bin/env python
import os
import abipy
import abipy.data

import abipy.gui.wxapps as wxapps

dirpath = os.path.join(abipy.data.dirpath, "runs", "data_si_ebands")
wxapps.wxapp_events(dirpath).MainLoop()



#!/usr/bin/env python
from __future__ import print_function, unicode_literals, division, absolute_import
import sys

from abipy.data import AbinitFilesGenerator


class MyGenerator(AbinitFilesGenerator):
    """This class generates the output files used in the unit tests and in the examples."""
    # Subclasses must define the following class attributes:
    # List of pseudos (basenames) in abipy/data/pseudos
    pseudos = ["14si.pspnc", "8o.pspnc"]

    # Mapping old_name --> new_name for the output files that must be preserved.
    files_to_save = {
        "out_DS1_DEN.nc": "sio2_DEN.nc",
        "out_DS2_GSR.nc": "sio2_kpath_GSR.nc",
        "out_DS3_SCR.nc": "sio2_SCR.nc",
        "out_DS3_EM1_NLF": "sio2_EM1_NLF",
        "out_DS3_EM1_LF": "sio2_EM1_LF",
        "out_DS3_EELF": "sio2_EELF",
    }


if __name__ == "__main__":
    sys.exit(MyGenerator().run())



#!/usr/bin/env python
r"""
Band structure plot
===================

This example shows how to plot a band structure
using the eigenvalues stored in the GSR file
produced at the end of the GS run.
"""
from abipy import abilab
import abipy.data as abidata
import os

dirpath = "."

gsr_paths = [os.path.join(dirpath, fname) for fname in ("mp-149_+0_GSR.nc", "mp-149_+2_GSR.nc")]
phbst_paths = [os.path.join(dirpath, fname) for fname in ("mp-149_+0_PHBST.nc", "mp-149_+2_PHBST.nc")]

nrows = len(gsr_paths)
ncols =  3
ax_mat, fig, plt = abilab.get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                              sharex=False, sharey=False, squeeze=False)

gsr_files = [abilab.abiopen(p) for p in gsr_paths]
phbst_files = [abilab.abiopen(p) for p in phbst_paths]

for i, (gsr, phbst) in enumerate(zip(gsr_files, phbst_files)):
    max_phfreq = phbst.phbands.maxfreq
    gsr.ebands.plot(ax=ax_mat[i, 0], with_gaps=True, max_phfreq=max_phfreq, show=False)
    gsr.ebands.interpolate(line_density=20).ebands_kpath.plot(ax=ax_mat[i, 1], with_gaps=True, max_phfreq=max_phfreq, show=False)
    phbst.phbands.plot(ax=ax_mat[i, 2], show=False)

plt.show()

# Here we use one of the GSR files shipped with abipy.
# Replace filename with the path to your GSR file or your WFK file.
#filename = abidata.ref_file("si_nscf_GSR.nc")

# Open the GSR file and extract the band structure.
# (alternatively one can use the shell and `abiopen.py OUT_GSR.nc -nb`
# to open the file in a jupyter notebook.
#with abiopen(filename) as ncfile:
#    ebands = ncfile.ebands

# Plot the band energies. Note that the labels for the k-points
# are found automatically in an internal database.
# Show fundamental and direct gaps.
#ebands.plot(with_gaps="fd", title="Silicon band structure")
#ebands.plot(with_gaps=True, title="Silicon band structure")

# Plot the BZ and the k-point path.
#ebands.kpoints.plot()



#!/usr/bin/env python
import wx
import numpy as np

from abipy.core.func1d import Function1D
from abipy.gui.awx.func1dframe import Func1dPlotFrame

app = wx.App()
func1d = Function1D.from_func(np.sin, np.arange(1,100,1))
frame = Func1dPlotFrame(None, func1d)
frame.Show()
app.MainLoop()



#!/usr/bin/env python
r"""
Raman Flow with independent-particle approximation (optic)
==========================================================

This script shows how to perform a Raman calculation with excitonic effects
included with the BSE formalism.
"""
from __future__ import division, print_function, unicode_literals, absolute_import

import sys
import os
import numpy as np

import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk


def build_flow(options):
    # Get the unperturbed structure.
    pseudos = abidata.pseudos("14si.pspnc")
    base_structure = abidata.structure_from_ucell("Si")

    ngkpt = [6, 6, 6]

    etas = [-.001, 0, +.001]
    ph_displ = np.reshape(np.zeros(3*len(base_structure)), (-1,3))
    ph_displ[0,:] = [+1, 0, 0]
    ph_displ[1,:] = [-1, 0, 0]

    # Build new structures by displacing atoms according to the phonon displacement
    # ph_displ (in cartesian coordinates). The Displacement is normalized so that
    # the maximum atomic diplacement is 1 Angstrom and then multiplied by eta.
    modifier = abilab.StructureModifier(base_structure)

    displaced_structures = modifier.displace(ph_displ, etas, frac_coords=False)

    # Initialize flow. Each workflow in the flow defines a complete BSE calculation for given eta.
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_","flow_")

    flow = flowtk.Flow(options.workdir, manager=options.manager)

    # Generate the different shifts to average
    ndiv = 1
    shift1D = np.arange(1, 2*ndiv+1, 2) / (2 * ndiv)
    all_shifts = [[x, y, z] for x in shift1D for y in shift1D for z in shift1D]

    for structure, eta in zip(displaced_structures, etas):
        for ishift,shift in enumerate(all_shifts):
            flow.register_work(raman_work(structure, pseudos, ngkpt, shift),
                               workdir="eta_" + str(eta) + "shift_" + str(ishift))

    return flow


def raman_work(structure, pseudos, ngkpt, shiftk):
    # Generate 3 different input files for computing optical properties with Optic.

    global_vars = dict(
        istwfk="*1",
        paral_kgb=0,
        ecut=8,
        nstep=200,
        diemac=12,
        ixc=7,
        chksymbreak=0,
        #accesswff=3
    )

    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=5)
    multi.set_vars(global_vars)
    multi.set_kmesh(ngkpt=ngkpt, shiftk=shiftk)

    # GS run
    multi[0].set_vars(
        tolvrs=1e-8,
        nband=20,
        nbdbuf=2,
    )

    # Note kptopt 2 in NSCF and DDK
    # In principle kptopt 2 is needed only in DDK.
    # one could do a first NSCF run with kptopt 1, reread with kptopt 2 and enter DDK...

    # NSCF run
    multi[1].set_vars(
       iscf=-2,
       nband=40,
       nbdbuf=5,
       kptopt=2,
       tolwfr=1.e-12,
    )

    # DDK along 3 directions
    # Third dataset : ddk response function along axis 1
    # Fourth dataset : ddk response function along axis 2
    # Fifth dataset : ddk response function along axis 3
    for idir in range(3):
        rfdir = 3 * [0]
        rfdir[idir] = 1

        multi[2 + idir].set_vars(
           iscf=-3,
           nband=40,
           nbdbuf=5,
           nstep=1,
           nline=0,
           prtwf=3,
           kptopt=2,
           nqpt=1,
           qpt=[0.0, 0.0, 0.0],
           rfdir=rfdir,
           rfelfd=2,
           tolwfr=1.e-12,
        )

    scf_inp, nscf_inp, ddk1, ddk2, ddk3 = multi.split_datasets()
    ddk_inputs = [ddk1, ddk2, ddk3]

    work = flowtk.Work()
    scf_t = work.register_scf_task(scf_inp)
    nscf_t = work.register_nscf_task(nscf_inp, deps={scf_t: "DEN"})

    ddk_nodes = []
    for inp in ddk_inputs:
        ddk_t = work.register_ddk_task(inp, deps={nscf_t: "WFK"})
        ddk_nodes.append(ddk_t)


    optic_input = abilab.OpticInput(
        broadening=0.002,     # Value of the smearing factor, in Hartree
        domega=0.0003,        # Frequency mesh.
        maxomega=0.3,
        scissor=0.000,        # Scissor shift if needed, in Hartree
        tolerance=0.002,      # Tolerance on closeness of singularities (in Hartree)
        num_lin_comp=6,       # Number of components of linear optic tensor to be computed
        lin_comp=(11, 12, 13, 22, 23, 33), # Linear coefficients to be computed (x=1, y=2, z=3)
        num_nonlin_comp=0     # Number of components of nonlinear optic tensor to be computed
        #nonlin_comp=(123, 222),
    )

    optic_t = flowtk.OpticTask(optic_input, nscf_node=nscf_t, ddk_nodes=ddk_nodes)
    work.register(optic_t)

    return work


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    #build_flow(options).plot_networkx(with_edge_labels=False, tight_layout=True)
    #dpi = 96
    #build_flow(options).graphviz_imshow(figsize=(800/dpi, 800/dpi), dpi=dpi)
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

import numpy as np
import pymatgen.io.abinit.netcdf as ionc

from monty.functools import lazy_property
from .xsf import *
from .visualizer import *


as_etsfreader = ionc.as_etsfreader


class ETSF_Reader(ionc.ETSF_Reader):
    """
    Provides high-level API to read data from netcdf files written
    folloing the ETSF-IO specifications described in :cite:`Caliste2008`
    """

    def read_structure(self):
        """
        Overrides the ``read_structure`` method so that we always return
        an instance of AbiPy |Structure| object
        """
        from abipy.core.structure import Structure
        return Structure.from_file(self.path)

    # Must overwrite implementation of pymatgen.io.abinit.netcdf
    # due to a possible bug introduced by initial whitespaces in symbol
    @lazy_property
    def chemical_symbols(self):
        """Chemical symbols char [number of atom species][symbol length]."""
        charr = self.read_value("chemical_symbols")
        symbols = []
        for v in charr:
            s = "".join(c.decode("utf-8") for c in v)
            # Strip to avoid possible whitespaces.
            symbols.append(s.strip())

        return symbols

    def read_string(self, varname):
        """
        Args:
            varname: Name of the variable
        """
        b = self.rootgrp.variables[varname][:]
        #print(type(b))
        import netCDF4
        try:
            value = netCDF4.chartostring(b)[()].decode('utf-8')
        except Exception:
            try:
                value = netCDF4.chartostring(b)[()]
            except Exception:
                try:
                    value = "".join(c for c in self.read_value(varname))
                except TypeError as exc:
                    value = "".join(c.decode("utf-8") for c in self.read_value(varname))

        return value.strip()

    def none_if_masked_array(self, arr):
        """Return None if arr is a MaskedArray else None."""
        return None if np.ma.is_masked(arr) else arr



# coding: utf-8
"""DDK.nc file."""
from __future__ import print_function, division, unicode_literals, absolute_import

import numpy as np
import pymatgen.core.units as units

from monty.string import marquee # is_string, list_strings,
from monty.functools import lazy_property
from monty.collections import dict2namedtuple
from abipy.core.func1d import Function1D
from abipy.core.mixins import AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.tools import gaussian, duck
from abipy.electrons.ebands import ElectronsReader
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt #, set_axlims


class DdksAnalyzer(object):
    """
    This object received three ddks files with derivatives along the 3 reduced directions.
    """

    def __init__(self, ddk_paths):
        # Open files here. The order of input ddk_paths does not matter as long as we can fill all ddks.
        self.ddks = ddks = 3 * [None]
        for path in ddk_paths:
            ddk = DdkFile(path)
            ddks[ddk.idir - 1] = ddk
        if any(ddk is None for ddk in ddks):
            raise ValueError("Cannot find 3 DDK files with different idir.")

        # Consistency check
        errors = []
        eapp = errors.append
        if any(ddk.structure != ddks[0].structure for ddk in ddks[1:]):
            eapp("Structures in DDK files do not agree with each other.")
        if any(ddk.kptopt != ddks[0].kptopt for ddk in ddks[1:]):
            eapp("Found different values of kptopt.")
        if any(ddk.ebands.kpoints != ddks[0].ebands.kpoints for ddk in ddks[1:]):
            eapp("Found different list of k-kpoints.")
        for aname in ("nsppol", "nspden", "nspinor"):
            if any(getattr(ddk.ebands, aname) != getattr(ddks[0].ebands, aname) for ddk in ddks[1:]):
                eapp("Found different value of %s" % aname)
        if errors:
            raise ValueError("\n".join(errors))

        # Get useful dimensions.
        eb0 = ddks[0].ebands
        self.nsppol, self.nspinor, self.nspden = eb0.nsppol, eb0.nspinor, eb0.nspden
        self.nband = eb0.nband
        self.kpoints = eb0.kpoints
        self.nkpt = len(eb0.kpoints)
        self.weights = eb0.kpoints.weights
        self.eigens = eb0.eigens

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for ddk in self.ddks:
            try:
                ddk.close()
            except Exception:
                pass

    def __str__(self):
        """String representation."""
        return self.to_string()

    def to_string(self, verbose=0):
        """String representation."""
        return "\n\n".join(ddk.to_string(verbose=verbose) for ddk in self.ddks)

    @lazy_property
    def v_skb(self):
        v_skb = np.empty((self.nsppol, self.nkpt, self.nband, 3))
        for i, ddk in enumerate(self.ddks):
            v_skb[:, :, :, i] = ddk.reader.read_ddk_diagonal()
        return v_skb

    #def get_averaged_v(self, isolevels):

    def get_doses(self, method="gaussian", step=0.1, width=0.2):
        """
        Compute the electronic DOS on a linear mesh.

        Args:
            method: String defining the method for the computation of the DOS.
            step: Energy step (eV) of the linear mesh.
            width: Standard deviation (eV) of the gaussian.

        Returns: |ElectronDos| object.
        """
        self.kpoints.check_weights()
        edos = self.ddks[0].ebands.get_edos(method=method, step=step, width=width)
        values = np.zeros((self.nsppol, nw))
        mesh = edos[0].mesh
        #vmod_skb = self.vskb
        if method == "gaussian":
            for spin in range(self.nsppol):
                for k, kpoint in enumerate(self.kpoints):
                    wk = kpoint.weight
                    for band in range(self.nband):
                        e = self.eigens[spin, k, band]
                        values[spin] += wk * vmod[spin, k, band] * gaussian(mesh, width, center=e)
        else:
            raise NotImplementedError("Method %s is not supported" % method)

        vdos_spin = [Function1D(mesh, values[spin]) for spin in range(self.nsppol)]
        vdos = 2 * Function1D(mesh, values[0]) if self.nsppol == 1 else vdos_spin[0] + vdos_spin[1]

        return dict2namedtuple(edos=edos, vdos=vdos, vdos_spin=vdos_spin)

    @add_fig_kwargs
    def plot_vdos(self, method="gaussian", step=0.1, width=0.2, **kwargs):
        """

        Return: |matplotlib-Figure|
        """
        import matplotlib.pyplot as plt
        fig, ax_mat = plt.subplots(nrows=2, ncols=1, sharex=True, sharey=False, squeeze=True)
        r = self.get_doses(method=method, step=step, width=width)
        #r.edos
        #r.vdos
        return fig

    @add_fig_kwargs
    def plot_ebands_with_doses(self, ebands_kpath, doses, ylims=None, **kwargs):
        """
        Plot band structure and doses

        Args:
            ebands_kpath:
            doses:
            ylims: Set the data limits for the x-axis in eV. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used

        Return: |matplotlib-Figure|
        """
        # Build grid plot.
        import matplotlib.pyplot as plt
        from matplotlib.gridspec import GridSpec
        fig = plt.figure()
        ncols = 3
        width_ratios = [2, 0.2, 0.2]
        gspec = GridSpec(1, ncols, width_ratios=width_ratios)
        gspec.update(wspace=0.05)

        ax_ebands = plt.subplot(gspec[0])
        ax_doses = []
        for i in range(2):
            ax = plt.subplot(gspec[i + 1], sharey=ax_ebands)
            ax_doses.append(ax)
            ax.grid(True)
            set_axlims(ax, ylims, "y")

        # Plot electron bands.
        ebands_kpath.plot(ax=ax_ebands, ylims=ylims, show=False)

        # Plot DOSes.
        #doses.edos.plot
        #vdos.edos.plot
        #ax.set_ylabel("")

        return fig

    # TODO
    #def plot_vfield(self, **kwargs):
    #def plot_v_on_isosurface(self, **kwargs):


class DdkFile(AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    File containing the DDK matrix elements for a single `idir` direction

    Usage example:

    .. code-block:: python

        with DdkFile("out_DDK.nc") as ddk:
            ddk.ebands.plot()
    """
    @classmethod
    def from_file(cls, filepath):
        """Initialize the object from a netcdf_ file."""
        return cls(filepath)

    def __init__(self, filepath):
        super(DdkFile, self).__init__(filepath)
        self.reader = DdkReader(filepath)

        # Get info on perturbation and k-point sampling.
        self.kptopt = self.reader.read_value("kptopt")
        #assert self.kptopt == 2
        pertcase = self.reader.read_value("pertcase")
        self.idir = ((pertcase - 1) % 3) + 1
        self.ipert = (pertcase - self.idir) // 3 + 1

    #@lazy_property
    #def dedk_idir
    #    return self.reader.read_value("dedk_bbmat_idir")

    def __str__(self):
        """String representation."""
        return self.to_string()

    def to_string(self, verbose=0):
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.ebands.to_string(with_structure=False, title="Electronic Bands"))
        app(marquee("DDK perturbation", mark="="))
        app("idir: {}, ipert: {}, kptopt: {}".format(self.idir, self.ipert, self.kptopt))

        if verbose > 1:
            app("")
            app(self.hdr.to_string(verbose=verbose, title="Abinit Header"))

        return "\n".join(lines)

    @lazy_property
    def ebands(self):
        """|ElectronBands| object."""
        return self.reader.read_ebands()

    @property
    def structure(self):
        """|Structure| object."""
        return self.ebands.structure

    @lazy_property
    def params(self):
        """:class:`OrderedDict` with parameters that might be subject to convergence studies."""
        od = self.get_ebands_params()
        return od

    def close(self):
        """Close the file."""
        self.reader.close()

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        yield self.structure.plot(show=False)
        #yield self.ebands.plot(show=False)

    def write_notebook(self, nbpath=None, title=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=title)

        nb.cells.extend([
            nbv.new_code_cell("ddk = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(ddk)"),
            nbv.new_code_cell("ddk.ebands.plot();"),
            #nbv.new_code_cell("ddk.ebands.kpoints.plot();"),
            nbv.new_code_cell("""\
if ddk.ebands.kpoints.is_ibz:
    ddk.ebands.get_edos().plot();"""),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class DdkReader(ElectronsReader):
    """
    This object reads the results stored in the DDK.nc file
    It provides helper function to access the most important quantities.
    """
    def __init__(self, filepath):
        super(DdkReader, self).__init__(filepath)
        nband_sk = self.read_nband_sk()
        if np.any(nband_sk != nband_sk[0, 0]):
            raise NotImplementedError("Found different number of bands per k-point, spin.\nnband_sk: %s\n" %
                    str(nband_sk))
        self.mband = nband_sk[0, 0]

    def read_ddk_diagonal(self):
        """
        Read the group velocities i.e the diagonal matrix elements.
        Return (nsppol, nkpt) |numpy-array| of real numbers.
        """
        var = self.read_variable("h1_matrix")
        vels = np.diagonal(var[:, :, :, :, :], axis1=2, axis=3)
        # Cartesian? Ha --> eV?
        return np.real(vels).copy() * (units.Ha_to_eV / units.bohr_to_ang)

    def read_ddk_skbb(self):
        return self.read_value("h1_matrix", cplx_mode="cplx") * (units.Ha_to_eV / units.bohr_to_ang)



"""
This module gathers the most important classes and helper functions used for scripting.
"""
from __future__ import print_function, division, unicode_literals

import sys
import os
import collections

####################
### Monty import ###
####################
from monty.os.path import which

#######################
### Pymatgen import ###
#######################
# Tools for unit conversion
import pymatgen.core.units as units
FloatWithUnit = units.FloatWithUnit
ArrayWithUnit = units.ArrayWithUnit

####################
### Abipy import ###
####################
from abipy.flowtk import Pseudo, PseudoTable, Mrgscr, Mrgddb, Mrggkk, Flow, Work, TaskManager, AbinitBuild, flow_main
from abipy.core.release import __version__, min_abinit_version
from abipy.core.globals import enable_notebook, in_notebook, disable_notebook
from abipy.core import restapi
from abipy.core.structure import (Lattice, Structure, StructureModifier, dataframes_from_structures,
  mp_match_structure, mp_search, cod_search)
from abipy.core.mixins import CubeFile
from abipy.core.func1d import Function1D
from abipy.core.kpoints import set_atol_kdiff
from abipy.abio.robots import Robot
from abipy.abio.inputs import AbinitInput, MultiDataset, AnaddbInput, OpticInput
from abipy.abio.abivars import AbinitInputFile
from abipy.abio.outputs import AbinitLogFile, AbinitOutputFile, OutNcFile, AboRobot #, CubeFile
from abipy.tools.printing import print_dataframe
from abipy.tools.notebooks import print_source, print_doc
from abipy.tools.plotting import get_ax_fig_plt, get_axarray_fig_plt, get_ax3d_fig_plt #, plot_array, ArrayPlotter
from abipy.abio.factories import *
from abipy.electrons.ebands import (ElectronBands, ElectronBandsPlotter, ElectronDos, ElectronDosPlotter,
    dataframe_from_ebands)
from abipy.electrons.gsr import GsrFile, GsrRobot
from abipy.electrons.psps import PspsFile
from abipy.electrons.ddk import DdkFile
from abipy.electrons.gw import SigresFile, SigresRobot
from abipy.electrons.bse import MdfFile, MdfRobot
from abipy.electrons.scissors import ScissorsBuilder
from abipy.electrons.scr import ScrFile
from abipy.electrons.denpot import (DensityNcFile, VhartreeNcFile, VxcNcFile, VhxcNcFile, PotNcFile,
    DensityFortranFile, Cut3dDenPotNcFile)
from abipy.electrons.fatbands import FatBandsFile
from abipy.electrons.optic import OpticNcFile, OpticRobot
from abipy.electrons.fold2bloch import Fold2BlochNcfile
from abipy.dfpt.phonons import (PhbstFile, PhbstRobot, PhononBands, PhononBandsPlotter, PhdosFile, PhononDosPlotter,
    PhdosReader, phbands_gridplot)
from abipy.dfpt.ddb import DdbFile, DdbRobot
from abipy.dfpt.anaddbnc import AnaddbNcFile, AnaddbNcRobot
from abipy.dfpt.gruneisen import GrunsNcFile
from abipy.dynamics.hist import HistFile, HistRobot
from abipy.waves import WfkFile
from abipy.eph.a2f import A2fFile, A2fRobot
from abipy.eph.sigeph import SigEPhFile, SigEPhRobot
from abipy.wannier90 import WoutFile, AbiwanFile, AbiwanRobot
from abipy.electrons.lobster import CoxpFile, ICoxpFile, LobsterDoscarFile, LobsterInput, LobsterAnalyzer

# Abinit Documentation.
from abipy.abio.abivars_db import get_abinit_variables, abinit_help, docvar


def _straceback():
    """Returns a string with the traceback."""
    import traceback
    return traceback.format_exc()

# Abinit text files. Use OrderedDict for nice output in show_abiopen_exc2class.
ext2file = collections.OrderedDict([
    (".abi", AbinitInputFile),
    (".in", AbinitInputFile),
    (".abo", AbinitOutputFile),
    (".out", AbinitOutputFile),
    (".log", AbinitLogFile),
    (".cif", Structure),
    ("POSCAR", Structure),
    (".cssr", Structure),
    (".cube", CubeFile),
    ("anaddb.nc", AnaddbNcFile),
    ("DEN", DensityFortranFile),
    (".psp8", Pseudo),
    (".pspnc", Pseudo),
    (".fhi", Pseudo),
    ("JTH.xml", Pseudo),
    (".wout", WoutFile),
    # Lobster files.
    ("COHPCAR.lobster", CoxpFile),
    ("COOPCAR.lobster", CoxpFile),
    ("ICOHPLIST.lobster", ICoxpFile),
    ("DOSCAR.lobster", LobsterDoscarFile),
])

# Abinit files require a special treatment.
abiext2ncfile = collections.OrderedDict([
    ("GSR.nc", GsrFile),
    ("DEN.nc", DensityNcFile),
    ("OUT.nc", OutNcFile),
    ("DDK.nc", DdkFile),
    ("VHA.nc", VhartreeNcFile),
    ("VXC.nc", VxcNcFile),
    ("VHXC.nc", VhxcNcFile),
    ("POT.nc", PotNcFile),
    ("WFK.nc", WfkFile),
    ("HIST.nc", HistFile),
    ("PSPS.nc", PspsFile),
    ("DDB", DdbFile),
    ("PHBST.nc", PhbstFile),
    ("PHDOS.nc", PhdosFile),
    ("SCR.nc", ScrFile),
    ("SIGRES.nc", SigresFile),
    ("GRUNS.nc", GrunsNcFile),
    ("MDF.nc", MdfFile),
    ("FATBANDS.nc", FatBandsFile),
    ("FOLD2BLOCH.nc", Fold2BlochNcfile),
    ("CUT3DDENPOT.nc", Cut3dDenPotNcFile),
    ("OPTIC.nc", OpticNcFile),
    ("A2F.nc", A2fFile),
    ("SIGEPH.nc", SigEPhFile),
    ("ABIWAN.nc", AbiwanFile),
])


def abiopen_ext2class_table():
    """
    Print the association table between file extensions and File classes.
    """
    from itertools import chain
    from tabulate import tabulate
    table = []

    for ext, cls in chain(ext2file.items(), abiext2ncfile.items()):
        table.append((ext, str(cls)))

    return tabulate(table, headers=["Extension", "Class"])


def abifile_subclass_from_filename(filename):
    """
    Returns the appropriate class associated to the given filename.
    """
    if os.path.basename(filename) == Flow.PICKLE_FNAME:
        return Flow

    from abipy.tools.text import rreplace
    for ext, cls in ext2file.items():
        # This to support gzipped files.
        if filename.endswith(".gz"): filename = rreplace(filename, ".gz", "", occurrence=1)
        if filename.endswith(ext): return cls

    ext = filename.split("_")[-1]
    try:
        return abiext2ncfile[ext]
    except KeyError:
        for ext, cls in abiext2ncfile.items():
            if filename.endswith(ext): return cls

    msg = ("No class has been registered for file:\n\t%s\n\nFile extensions supported:\n\n%s" %
        (filename, abiopen_ext2class_table()))
    raise ValueError(msg)


def dir2abifiles(top, recurse=True):
    """
    Analyze the filesystem starting from directory `top` and
    return an ordered dictionary mapping the directory name to the list
    of files supported by ``abiopen`` contained within that directory.
    If not ``recurse``, children directories are not analyzed.
    """
    dl = collections.defaultdict(list)

    if recurse:
        for dirpath, dirnames, filenames in os.walk(top):
            for f in filenames:
                path = os.path.join(dirpath, f)
                if not isabifile(path): continue
                dl[dirpath].append(path)
    else:
        for f in os.listdir(top):
            path = os.path.join(top, f)
            if not isabifile(path): continue
            dl[top].append(path)

    return collections.OrderedDict([(k, dl[k]) for k in sorted(dl.keys())])


def isabifile(filepath):
    """
    Return True if `filepath` can be opened with ``abiopen``.
    """
    try:
        abifile_subclass_from_filename(filepath)
        return True
    except ValueError:
        return False


def abiopen(filepath):
    """
    Factory function that opens any file supported by abipy.
    File type is detected from the extension

    Args:
        filepath: string with the filename.
    """
    if os.path.basename(filepath) == "__AbinitFlow__.pickle":
        return Flow.pickle_load(filepath)

    # Handle old output files produced by Abinit.
    import re
    outnum = re.compile(r".+\.out[\d]+")
    abonum = re.compile(r".+\.abo[\d]+")
    if outnum.match(filepath) or abonum.match(filepath):
        return AbinitOutputFile.from_file(filepath)

    if os.path.basename(filepath) == "log":
        # Assume Abinit log file.
        return AbinitLogFile.from_file(filepath)

    cls = abifile_subclass_from_filename(filepath)
    return cls.from_file(filepath)


def display_structure(obj, **kwargs):
    """
    Use Jsmol to display a structure in the jupyter notebook.
    Requires `nbjsmol` notebook extension installed on the local machine.
    Install it with `pip install nbjsmol`. See also https://github.com/gmatteo/nbjsmol.

    Args:
        obj: Structure object or file with a structure or python object with a `structure` attribute.
        kwargs: Keyword arguments passed to `nbjsmol_display`
    """
    try:
        from nbjsmol import nbjsmol_display
    except ImportError as exc:
        raise ImportError(str(exc) +
                          "\ndisplay structure requires nbjsmol package\n."
                          "Install it with `pip install nbjsmol.`\n"
                          "See also https://github.com/gmatteo/nbjsmol.")

    # Cast to structure, get string with cif data and pass it to nbjsmol.
    structure = Structure.as_structure(obj)
    return nbjsmol_display(structure.to(fmt="cif"), ext=".cif", **kwargs)


def software_stack():
    """
    Import all the hard dependencies. Returns ordered dict: package --> string with version info.
    """
    import platform
    system, node, release, version, machine, processor = platform.uname()
    # These packages are required
    import numpy, scipy, netCDF4, pymatgen, apscheduler, pydispatch, yaml

    d = collections.OrderedDict([
        ("system", system),
        ("python_version", platform.python_version()),
        ("numpy", numpy.version.version),
        ("scipy", scipy.version.version),
        ("netCDF4", netCDF4.__version__),
        ("apscheduler", apscheduler.version),
        ("pydispatch", pydispatch.__version__),
        ("yaml", yaml.__version__),
        ("pymatgen", pymatgen.__version__),
    ])

    # Optional but strongly suggested.
    #try:
    #    import matplotlib
    #    d["matplotlib"] = "%s (backend: %s)" % (matplotlib.__version__, matplotlib.get_backend())
    #except ImportError:
    #    pass

    return d


def abicheck(verbose=0):
    """
    This function tests if the most important ABINIT executables
    can be found in $PATH and whether the python modules needed
    at run-time can be imported. Return string with error messages, empty if success.
    """
    from monty.termcolor import cprint
    err_lines = []
    app = err_lines.append

    try:
        manager = TaskManager.from_user_config()
    except Exception:
        manager = None
        app(_straceback())

    # Get info on the Abinit build.
    from abipy.core.testing import cmp_version
    from abipy.flowtk import PyFlowScheduler

    if manager is not None:
        cprint("AbiPy Manager:\n%s\n" % str(manager), color="green")
        build = AbinitBuild(manager=manager)
        if not build.has_netcdf: app("Abinit executable does not support netcdf")
        cprint("Abinitbuild:\n%s" % str(build), color="magenta")
        if verbose: cprint(str(build.info), color="magenta")
        print()
        if not cmp_version(build.version, min_abinit_version, op=">="):
            app("Abipy requires Abinit version >= %s but got %s" % (min_abinit_version, build.version))

    # Get info on the scheduler.
    try:
        scheduler = PyFlowScheduler.from_user_config()
        cprint("Abipy Scheduler:\n%s\n" % str(scheduler), color="yellow")
    except Exception as exc:
        app(_straceback())

    from tabulate import tabulate
    try:
        d = software_stack()
        cprint("Installed packages:", color="blue")
        cprint(tabulate(list(d.items()), headers=["Package", "Version"]), color="blue")
        print()
    except ImportError:
        app(_straceback())

    return "\n".join(err_lines)


def abipy_logo1():
    """http://www.text-image.com/convert/pic2ascii.cgi"""
    return r"""
                 `:-                                                               -:`
         --`  .+/`                              `                                  `/+.  .-.
   `.  :+.   /s-                   `yy         .yo                                   -s/   :+. .`
 ./.  +o`   /s/           `-::-`   `yy.-::-`   `:-    .:::-`   -:`     .:`            /s/   :s- ./.
.o.  /o:   .oo.         .oyo++syo. `yyys++oys. -ys  -syo++sy+` sy-     +y:            .oo-   oo` `o.
++   oo.   /oo          yy-    -yy `yy:    .yy`-ys .ys`    /yo sy-     +y:             oo/   /o:  ++
+/   oo`   /oo         `yy.    .yy` yy.    `yy`-ys :ys     :yo oy/     oy:             +o/   :o:  /o
-/   :+.   -++`         -sy+::+yyy` .sy+::+yy- -ys :yys/::oys. `oyo::/syy:            `++-   /+.  /:
 --  `//    /+-           -/++/-//    -/++/-   `+: :yo:/++/.     .:++/:oy:            -+/   `+-  --
  `.`  -:    :/`                                   :yo                 +y:           `/:`  `:. `.`
        `..   .:.                                   .`                 `.           .:.  `..
                ...                                                               ...
"""


def abipy_logo2():
    """http://www.text-image.com/convert/pic2ascii.cgi"""
    return r"""
MMMMMMMMMMMMMMMMNhdMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMdhmMMMMMMMMMMMMMMM
MMMMMMMMMddNMMmoyNMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMNMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMNyomMMmhmMMMMMMMM
MMMmmMMhomMMMy/hMMMMMMMMMMMMMMMMMMMN::MMMMMMMMMm:oMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMd+yMMMhomMmmMMM
MmsmMMs+NMMMy+yMMMMMMMMMMMNhyyhmMMMN::mhyyhmMMMNhdMMMMmhyydNMMMdyNMMMMMmyNMMMMMMMMMMMMy+yMMMh+dMmsmM
m+mMMy+hMMMd++mMMMMMMMMMm+:+ss+:+mMN:::/ss+:+mMd:/MMd/:+so/:oNM/:dMMMMMo:yMMMMMMMMMMMMm++dMMMo+NMN+m
osMMMo+mMMMy+oMMMMMMMMMM::dMMMMd:/MN::hMMMMd::Nd:/Mm:/NMMMMy:oM/:dMMMMMo:yMMMMMMMMMMMMMo+yMMMy+hMMso
oyMMMooNMMMyooMMMMMMMMMN::mMMMMm::NM::mMMMMN::Nd:/Mh:/MMMMMh:+Mo:yMMMMM+:yMMMMMMMMMMMMMooyMMMyohMMyo
dyMMMysmMMMdooNMMMMMMMMMd/:oyys:::NMd/:oyys::dMd:/Mh::/shyo:/mMN+:+yhs/::yMMMMMMMMMMMMNoodMMMysmMMyh
MddMMNyhMMMMysdMMMMMMMMMMMdyooydssMMMMdysoydMMMNsyMh:+hsosydMMMMMmysosho:yMMMMMMMMMMMMdsyMMMNydMMddM
MMNmNMMddMMMMhyNMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMh:oMMMMMMMMMMMMMMMMMo:yMMMMMMMMMMMNyhMMMNhmMNmNMM
MMMMMMMMNmmMMMmhmMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMmNMMMMMMMMMMMMMMMMMNdMMMMMMMMMMMmhmMMNmmMMMMMMMM
MMMMMMMMMMMMMMMMmmNMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMNmmMMMMMMMMMMMMMMM
"""


def abipy_logo3():
    """http://www.text-image.com/convert/pic2ascii.cgi"""
    return r"""\
             `-.                                                  `--`
      -:. `//`              `/.       ::                           `+: `-:`
 --``+:  `o+        `.--.`  -y/.--.`  ::   `---`  `-     ..         `o+  `o-`--
:/  /o   /o-       -oo/:+s/ -yy+:/os- ss .oo/:/s+`:y.    yo          :o:  :o. /:
o-  o/   oo`       ss    /y..y/    ss ss +y`   -y::y-    yo          -o/  .o- -o
:-  //   /+.       :s+--/sy- +s/--+s: ss oyo:-:so``os:-:oyo          -+:  -+. -/
 -` `/.  `/:        `-:::-:`  `-::-`  -- oy-:::.    .:::-yo          //`  :- `-
   `  ..` `:-                            :+              /:         --` `-` `
            `.`                                                   ..`
"""



# coding: utf-8
"""Helper functions."""
from __future__ import print_function, division, absolute_import, unicode_literals

from .devtools import *
from .iotools  import *
from .numtools import *
from .text import *


class NoDefaultProvided(object):
    pass


def hasattrd(obj, name):
    """
    The arguments are an object and a string.
    The result is True if the string is the name of one of the object’s attributes, False if not.
    Unlike the builtin hasattr, hasattrd supports dot notation e.g. hasattr(int, "__class__.__name__")
    (This is implemented by calling getattrd(object, name) and seeing whether it raises an exception or not.)
    """
    try:
        getattrd(obj, name)
        return True
    except AttributeError:
        return False


def getattrd(obj, name, default=NoDefaultProvided):
    """
    Same as getattr(), but allows dot notation lookup e.g. getattrd(obj, "a.b")

    Raises: AttributeError if ``name`` is not found and ``default`` is not given.

    Discussed in: http://stackoverflow.com/questions/11975781
    """
    from functools import reduce
    try:
        return reduce(getattr, name.split("."), obj)
    except AttributeError:
        if default is not NoDefaultProvided:
            return default
        raise



#!/usr/bin/env python
"""
Benchmark paral_kgb=1 algorithm with wfoptalg in [default, 1].
default correspongs to the lobpcg algorithm, 1 enables the Chebyschev solver.
"""
from __future__ import division, print_function, unicode_literals, absolute_import

import sys
import operator
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata

from itertools import product
from functools import reduce
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_input(paw=False):
    """
    Build and return an input file for GS calculations with paral_kgb=1
    """
    pseudos = abidata.pseudos("14si.pspnc", "8o.pspnc")
              #if not paw else
              #abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")
    structure = abidata.structure_from_ucell("SiO2-alpha")

    inp = abilab.AbinitInput(structure, pseudos)
    inp.set_kmesh(ngkpt=[4, 4, 4], shiftk=[0,0,0.5])

    # Global variables
    ecut = 40
    inp.set_vars(
        ecut=ecut,
        pawecutdg=ecut*4 if paw else None,
        nsppol=1,
        nband=32,
        paral_kgb=1,
        kptopt=3,
        istwfk="*1",
        timopt=-1,
        chksymbreak=0,
        nsym=1,
        prtwf=0,
        prtden=0,
        tolvrs=1e-8,
        nstep=20,
    )

    return inp


def build_flow(options):
    template = make_input()

    # Get the list of possible parallel configurations from abinit autoparal.
    #max_ncpus, min_eff = options.max_ncpus, options.min_eff
    #print("Getting all autoparal configurations up to max_ncpus: ",max_ncpus," with efficiency >= ",min_eff)
    #pconfs = template.abiget_autoparal_pconfs(max_ncpus, autoparal=1, verbose=options.verbose)
    #if options.verbose: print(pconfs)

    # Processor distribution.
    pconfs = [
      dict(npkpt=64, npband=1, npfft=2), # 128
      dict(npkpt=64, npband=2, npfft=2), # 256
      dict(npkpt=64, npband=2, npfft=4), # 512
      dict(npkpt=64, npband=4, npfft=4), # 1024
    ]

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)

    for wfoptalg in [None, 1]:
        work = flowtk.Work()
        for conf, omp_threads in product(pconfs, options.omp_list):
            #if not options.accept_conf(conf, omp_threads): continue
            mpi_procs = omp_threads * reduce(operator.mul, conf.values(), 1)

            manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
            inp = template.new_with_vars(conf, wfoptalg=wfoptalg)
            work.register_scf_task(inp, manager=manager)

        print("Found %d configurations" % len(work))
        flow.register_work(work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import print_function, division, unicode_literals, absolute_import

import wx

from monty.collections import AttrDict
from abipy.gui import awx
from abipy.electrons import ScissorsBuilder


class ScissorsParamsDialog(wx.Dialog):
    """
    This dialog prompts the user to enter the parameters for the
    generations of the Scissors operator.
    """
    #TODO: See
    #http://www.blog.pythonlibrary.org/2012/05/05/wxpython-adding-and-removing-widgets-dynamically/
    # for how to add and remove widgets dinamically.
    def __init__(self, parent, nsppol, num_domains, e0min, e0max, **kwargs):
        """
        Args:
            parent:
                parent window.
            nsppol:
                Number of spins.
            num_domains:
                Number of domains for the scissors.
            e0min:
                Minimum KS energy [eV]
            e0max:
                Maximum KS energy [eV]
        """
        assert nsppol == 1

        super(ScissorsParamsDialog, self).__init__(parent, -1, **kwargs)
        self.SetTitle("Select scissors parameters")

        main_sizer = wx.BoxSizer(wx.VERTICAL)

        # Construct a panel for each spin.
        self.panel_spin = []
        for spin in range(nsppol):
            panel = ScissorsParamsPanel(self, num_domains, e0min, e0max)
            self.panel_spin.append(panel)

            main_sizer.Add(panel, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)

        hbox = wx.BoxSizer(wx.HORIZONTAL)

        ok_button = wx.Button(self, wx.ID_OK, label='Ok')
        close_button = wx.Button(self, wx.ID_CANCEL, label='Cancel')

        hbox.Add(ok_button)
        hbox.Add(close_button, flag=wx.LEFT, border=5)

        main_sizer.Add(hbox, flag=wx.ALIGN_CENTER | wx.TOP | wx.BOTTOM, border=10)

        self.SetSizerAndFit(main_sizer)

    def GetScissorBuilderParams(self):
        """Returns a `AttrDict` with the parameters required to build the scissors."""
        domains_spin, bounds_spin = [], []
        for panel in self.panel_spin:
            p = panel.GetParams()
            domains_spin.append(p.domains)
            bounds_spin.append(p.bounds)

        return AttrDict(
            domains_spin=domains_spin,
            bounds_spin=bounds_spin,
        )


class ScissorsParamsPanel(awx.Panel):
    """
    Panel with controls allowing the user to specify the
    Scissors parameters for a single spin.
    """
    def __init__(self, parent, num_domains, e0min, e0max, **kwargs):
        """
        Args:
            parent:
                Parent window.
            num_domains:
                Number of domains.
                e0min:
                    Minimum KS energy [eV]
                e0max:
                    Maximum KS energy [eV]
        """
        awx.Panel.__init__(self, parent, -1, **kwargs)

        self.bound_choices = ["c", "h"]

        # Enlarge the range used in SpinCtrlDouble.
        self.emin = e0min - 0.1 * abs(e0min)
        self.emax = e0max + 0.1 * abs(e0max)
        self.num_domains = num_domains

        self.BuildUi()

    def BuildUi(self):
        main_sizer = wx.BoxSizer(wx.VERTICAL)

        # List of controls. Each control is associated to a particular domain.
        self._control_domains = []

        for numdom in range(self.num_domains):
            static_boxsizer = self._AddControlBox(numdom)
            main_sizer.Add(static_boxsizer, 1, wx.EXPAND, 5)

        self.SetSizer(main_sizer)
        self.Fit()

    def _AddControlBox(self, numdom):
        static_boxsizer = wx.StaticBoxSizer(wx.StaticBox(self, -1, "Domain #%d" % numdom), wx.VERTICAL)

        vsizer1 = wx.BoxSizer(wx.VERTICAL)

        flexgrid_sizer1 = wx.FlexGridSizer(0, 4, 0, 0)
        flexgrid_sizer1.AddGrowableCol(1)
        flexgrid_sizer1.SetFlexibleDirection(wx.HORIZONTAL)
        flexgrid_sizer1.SetNonFlexibleGrowMode(wx.FLEX_GROWMODE_SPECIFIED)

        stext1 = wx.StaticText(self, -1, "Emin:", wx.DefaultPosition, wx.DefaultSize, 0)
        stext1.Wrap(-1)
        flexgrid_sizer1.Add(stext1, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)

        min_spinctrl = wx.SpinCtrlDouble(self, id=-1, value=str(self.emin), min=self.emin, max=self.emax, inc=0.1)
        flexgrid_sizer1.Add(min_spinctrl, 1, wx.ALL, 5)

        stext2 = wx.StaticText(self, -1, "Bound_type:", wx.DefaultPosition, wx.DefaultSize, 0)
        stext2.Wrap(-1)
        flexgrid_sizer1.Add(stext2, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)

        minbound_choice = wx.Choice(self, -1, wx.DefaultPosition, wx.DefaultSize, self.bound_choices, 0)
        minbound_choice.SetSelection(0)
        flexgrid_sizer1.Add(minbound_choice, 0, wx.ALL, 5)

        stext3 = wx.StaticText(self, -1, "Emax:", wx.DefaultPosition, wx.DefaultSize, 0)
        stext3.Wrap(-1)
        flexgrid_sizer1.Add(stext3, 0, wx.TOP | wx.BOTTOM | wx.LEFT | wx.ALIGN_CENTER_HORIZONTAL, 5)

        max_spinctrl = wx.SpinCtrlDouble(self, id=-1, value=str(self.emax), min=self.emin, max=self.emax, inc=0.1)
        flexgrid_sizer1.Add(max_spinctrl, 1, wx.ALL, 5)

        stext4 = wx.StaticText(self, -1, "Bound_type:", wx.DefaultPosition, wx.DefaultSize, 0)
        stext4.Wrap(-1)
        flexgrid_sizer1.Add(stext4, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)

        maxbound_choice = wx.Choice(self, -1, wx.DefaultPosition, wx.DefaultSize, self.bound_choices, 0)
        maxbound_choice.SetSelection(0)
        flexgrid_sizer1.Add(maxbound_choice, 0, wx.ALL, 5)

        vsizer1.Add(flexgrid_sizer1, 1, wx.EXPAND, 5)

        static_boxsizer.Add(vsizer1, 1, wx.EXPAND, 5)

        class Control(AttrDict):
            """Helper class that faciliates the extraction of the parameters in GetParams"""
            def get_eminmax(self):
                return [float(self.min_spinctrl.GetValue()),
                        float(self.max_spinctrl.GetValue())]

            def get_bound(self):
                return [self.minbound_choice.GetStringSelection(),
                        self.maxbound_choice.GetStringSelection()]

        control = Control(
            min_spinctrl=min_spinctrl,
            minbound_choice=minbound_choice,
            max_spinctrl=max_spinctrl,
            maxbound_choice=maxbound_choice,
        )

        self._control_domains.append(control)

        return static_boxsizer

    def GetParams(self):
        """"
        Return a AttrDict with all the parameters needed to
        generate the scissors operator
        """
        domains = [c.get_eminmax() for c in self._control_domains]
        bounds = [c.get_bound() for c in self._control_domains]

        return AttrDict(
            domains=domains,
            bounds=bounds,
        )


class ScissorsBuilderFrame(awx.Frame):
    def __init__(self, parent, filepath, **kwargs):
        super(ScissorsBuilderFrame, self).__init__(parent, -1, **kwargs)

        # Instanciate the object that will perform most of the work.
        self.scissors_builder = ScissorsBuilder.from_file(filepath)

        self.BuildUi()

    @property
    def nsppol(self):
        """Number of spins."""
        return self.scissors_builder.nsppol

    @property
    def e0min(self):
        """Minimum KS energy (over spins)"""
        return self.scissors_builder.e0min

    @property
    def e0max(self):
        """Maximum KS energy (over spins)"""
        return self.scissors_builder.e0max

    def BuildUi(self):
        main_sizer = wx.BoxSizer(wx.HORIZONTAL)

        stext = wx.StaticText(self, -1, "Numer of domains:", wx.DefaultPosition, wx.DefaultSize, 0)
        stext.Wrap(-1)
        main_sizer.Add(stext, 0, wx.TOP | wx.BOTTOM | wx.LEFT, 5)

        self.numdomains_ctrl = wx.SpinCtrl(self, -1, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize,
                                           wx.SP_ARROW_KEYS, 1, 100, 1)
        main_sizer.Add(self.numdomains_ctrl, 0, wx.ALL, 5)

        startfit_button = wx.Button(self, -1, "Start Fit", wx.DefaultPosition, wx.DefaultSize, 0)
        startfit_button.Bind(wx.EVT_BUTTON, self.OnStartFitButton)
        main_sizer.Add(startfit_button, 0, wx.ALL, 5)

        self.SetSizerAndFit(main_sizer)

    def OnStartFitButton(self, event):
        """Callback that start the generation of the scissors."""
        num_domains = int(self.numdomains_ctrl.GetValue())

        build_dialog = ScissorsParamsDialog(self, self.nsppol, num_domains, self.e0min, self.e0max)

        if build_dialog.ShowModal() == wx.ID_OK:
            # Get the parameters of the scissors.
            p = build_dialog.GetScissorBuilderParams()
            #logger.info("scissor params", p)

            # Try the fit.
            fitok = True
            try:
                self.scissors_builder.build(p.domains_spin, p.bounds_spin)
            except:
                awx.showErrorMessage(self)
                fitok = False

            # Plot the results and save the current contents in the file.
            if fitok:
                self.scissors_builder.plotfit()

                if awx.askUser(self, "Save data?"):
                    saveFileDialog = wx.FileDialog(self, "Save pickle file", "", "",
                                                   "Pickle files (*.pickle)|*.pickle",
                                                   wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)

                    if saveFileDialog.ShowModal() == wx.ID_OK:
                        savefile = saveFileDialog.GetPath()
                        #logger.info("About to save in %s" % savefile)
                        self.scissors_builder.save_data(savefile)

                    saveFileDialog.Destroy()

        build_dialog.Destroy()


def wxapp_scissors(filepath):
    app = awx.App()
    ScissorsBuilderFrame(None, filepath).Show()
    return app





# coding: utf-8
"""Decorators."""
from __future__ import print_function, division, unicode_literals, absolute_import

from functools import wraps


def return_straceback_ifexc(func):
    """
    Decorator for functions that are supposed to return a string for logging purposes (e.g. str)
    Instead of raising an exception, the decorated function returns a string with the
    traceback so that execution can continue.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            import traceback
            return traceback.format_exc()
    return wrapper



#!/usr/bin/env python
import abipy.data as abidata

from abipy.gui.wxapps import wxapp_phbstviewer

phbst_filepath = abidata.ref_file("trf2_5.out_PHBST.nc")

wxapp_phbstviewer(phbst_filepath).MainLoop()



# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

print(
    "You mean the perl language i.e. the only language that looks the same before and after RSA encryption?\n"
    "No way, use this instead\n")

import this



"""Integration tests for phonon flows."""
from __future__ import print_function, division, unicode_literals, absolute_import

import pytest
import numpy as np
import numpy.testing.utils as nptu
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk

from abipy.core.testing import has_matplotlib


def scf_ph_inputs(tvars=None):
    """
    This function constructs the input files for the phonon calculation:
    GS input + the input files for the phonon calculation.
    """
    # Crystalline AlAs: computation of the second derivative of the total energy
    structure = abidata.structure_from_ucell("AlAs")

    # List of q-points for the phonon calculation (4,4,4) mesh.
    qpoints = [
             0.00000000E+00,  0.00000000E+00,  0.00000000E+00,
             2.50000000E-01,  0.00000000E+00,  0.00000000E+00,
             5.00000000E-01,  0.00000000E+00,  0.00000000E+00,
             2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
             5.00000000E-01,  2.50000000E-01,  0.00000000E+00,
            -2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
             5.00000000E-01,  5.00000000E-01,  0.00000000E+00,
            -2.50000000E-01,  5.00000000E-01,  2.50000000E-01,
            ]

    qpoints = np.reshape(qpoints, (-1, 3))

    # Global variables used both for the GS and the DFPT run.
    global_vars = dict(nband=4,
                       ecut=3.0,
                       ngkpt=[4, 4, 4],
                       shiftk=[0, 0, 0],
                       tolvrs=1.0e-6,
                       paral_kgb=0 if tvars is None else tvars.paral_kgb,
                    )

    multi = abilab.MultiDataset(structure=structure, pseudos=abidata.pseudos("13al.981214.fhi", "33as.pspnc"),
                               ndtset=1 + len(qpoints))

    multi.set_vars(global_vars)

    for i, qpt in enumerate(qpoints):
        # Response-function calculation for phonons.
        multi[i+1].set_vars(
            nstep=20,
            rfphon=1,        # Will consider phonon-type perturbation
            nqpt=1,          # One wavevector is to be considered
            qpt=qpt,         # This wavevector is q=0 (Gamma)
            kptopt=3,
            )

            #rfatpol   1 1   # Only the first atom is displaced
            #rfdir   1 0 0   # Along the first reduced coordinate axis
            #kptopt   2      # Automatic generation of k points, taking

    # Split input into gs_inp and ph_inputs
    return multi.split_datasets()


def itest_phonon_flow(fwp, tvars):
    """
    Create a flow for phonon calculations:

        1) One work for the GS run.

        2) nqpt works for phonon calculations. Each work contains
           nirred tasks where nirred is the number of irreducible phonon perturbations
           for that particular q-point.
    """
    all_inps = scf_ph_inputs(tvars)
    scf_input, ph_inputs = all_inps[0], all_inps[1:]

    ph_ngqpt = (2, 2, 2)
    flow = flowtk.PhononFlow.from_scf_input(fwp.workdir, scf_input, ph_ngqpt=ph_ngqpt, with_becs=False)

    scheduler = flow.make_scheduler()
    assert scheduler.start() == 0
    flow.check_status(show=True)
    assert all(work.finalized for work in flow)
    assert flow.all_ok

    with flow.open_final_ddb() as ddb:
        ddb_path = ddb.filepath
        ddb.to_string(verbose=2)
        assert len(ddb.structure) == 2
        #assert ddb.qpoints.frac_coords

    # Test PhononTask inspect method
    ph_task = flow[1][0]
    if has_matplotlib():
        assert ph_task.inspect(show=False)
    # Test get_results
    #ph_task.get_results()

    # Build new work with Anaddb tasks.
    # Construct a manager with mpi_procs==1 since anaddb do not support mpi_procs > 1 (except in elphon)
    shell_manager = fwp.manager.to_shell_manager(mpi_procs=1)
    awork = flowtk.Work(manager=shell_manager)

    # Phonons bands and DOS with gaussian method
    anaddb_input = abilab.AnaddbInput.phbands_and_dos(
        scf_input.structure, ngqpt=ph_ngqpt, ndivsm=5, nqsmall=10, dos_method="gaussian: 0.001 eV")

    atask = flowtk.AnaddbTask(anaddb_input, ddb_node=ddb_path, manager=shell_manager)
    awork.register(atask)

    # Phonons bands and DOS with tetrahedron method
    anaddb_input = abilab.AnaddbInput.phbands_and_dos(
        scf_input.structure, ngqpt=ph_ngqpt, ndivsm=5, nqsmall=10, dos_method="tetra")

    atask = flowtk.AnaddbTask(anaddb_input, ddb_node=ddb_path, manager=shell_manager)
    awork.register(atask)

    flow.register_work(awork)
    flow.allocate()
    flow.build()

    for i, atask in enumerate(awork):
        atask.history.info("About to run anaddb task: %d", i)
        atask.start_and_wait()
        assert atask.status == atask.S_DONE
        atask.check_status()
        assert atask.status == atask.S_OK

        # These output files should be produced in the task workdir.
        # Actually they should be in outdir but anaddb uses different conventions.
        assert len(atask.wdir.list_filepaths(wildcard="*PHBST.nc")) == 1
        assert len(atask.wdir.list_filepaths(wildcard="*PHDOS.nc")) == 1

    #assert flow.validate_json_schema()


def itest_phonon_restart(fwp):
    """Test the restart of phonon calculations with the scheduler."""
    # Crystalline AlAs: computation of the second derivative of the total energy
    structure = abidata.structure_from_ucell("AlAs")

    # List of q-points for the phonon calculation (4,4,4) mesh.
    qpoints = np.reshape([
        0.00000000E+00,  0.00000000E+00,  0.00000000E+00,
        2.50000000E-01,  0.00000000E+00,  0.00000000E+00,
        #5.00000000E-01,  0.00000000E+00,  0.00000000E+00,  # XXX Uncomment this line to test restart from 1DEN
                                                            # Too long --> disabled
    ], (-1, 3))

    # Global variables used both for the GS and the DFPT run.
    global_vars = dict(
        nband=4,
        ecut=3.0,
        ngkpt=[4, 4, 4],
        shiftk=[0, 0, 0],
        tolvrs=1.0e-5,
   )

    multi = abilab.MultiDataset(structure=structure, pseudos=abidata.pseudos("13al.981214.fhi", "33as.pspnc"),
                                ndtset=1 + len(qpoints))

    multi.set_vars(global_vars)

    for i, qpt in enumerate(qpoints):
        # Response-function calculation for phonons.
        multi[i+1].set_vars(
            rfphon=1,        # Will consider phonon-type perturbation.
            nqpt=1,          # One wavevector is to be considered.
            qpt=qpt,         # q-wavevector.
            kptopt=3,
            nstep=5,         # This is to trigger the phonon restart.
        )
        #rfatpol   1 1   # Only the first atom is displaced
        #rfdir   1 0 0   # Along the first reduced coordinate axis
        #kptopt   2      # Automatic generation of k points, taking

                                                           # i == 0 --> restart from WFK
        if i == 1: multi[i+1].set_vars(prtwf=-1, nstep=5)  # Restart with WFK and smart- io.
        if i == 2: multi[i+1].set_vars(prtwf=0, nstep=8)   # Restart from 1DEN. Too long --> disabled.

    all_inps = multi.split_datasets()
    scf_input, ph_inputs = all_inps[0], all_inps[1:]

    flow = flowtk.phonon_flow(fwp.workdir, scf_input, ph_inputs, manager=fwp.manager)
    flow.set_garbage_collector()

    for task in flow.iflat_tasks():
        task.manager.qadapter.max_num_launches = 20

    assert flow.make_scheduler().start() == 0

    flow.check_status(show=True, verbose=1)
    assert all(work.finalized for work in flow)
    assert flow.all_ok

    assert sum(task.num_restarts for task in flow.iflat_tasks()) > 0



# coding: utf-8
"""This module ..."""
from __future__ import print_function, division, unicode_literals, absolute_import

import abc
import os
import six
import collections
import tempfile
import pickle

from time import ctime
from monty.os.path import which
from monty.termcolor import cprint
from monty.string import is_string
from monty.functools import lazy_property
from abipy.flowtk.netcdf import NetcdfReader, NO_DEFAULT


__all__ = [
    "AbinitNcFile",
    "Has_Structure",
    "Has_ElectronBands",
    "Has_PhononBands",
    "NotebookWriter",
    "Has_Header",
]

@six.add_metaclass(abc.ABCMeta)
class BaseFile(object):
    """
    Abstract base class defining the methods that must be implemented
    by the concrete classes representing the different files produced by ABINIT.
    """
    def __init__(self, filepath):
        self._filepath = os.path.abspath(filepath)

        # Save stat values
        stat = os.stat(filepath)
        self._last_atime = stat.st_atime
        self._last_mtime = stat.st_mtime
        self._last_ctime = stat.st_ctime

    def __repr__(self):
        return "<%s, %s>" % (self.__class__.__name__, self.relpath)

    @classmethod
    def from_file(cls, filepath):
        """Initialize the object from a string."""
        if isinstance(filepath, cls):
            return filepath

        #print("Perhaps the subclass", cls, "must redefine the classmethod from_file.")
        return cls(filepath)

    @property
    def filepath(self):
        """Absolute path of the file."""
        return self._filepath

    @property
    def relpath(self):
        """Relative path."""
        try:
            return os.path.relpath(self.filepath)
        except OSError:
            # current working directory may not be defined!
            return self.filepath

    @property
    def basename(self):
        """Basename of the file."""
        return os.path.basename(self.filepath)

    @property
    def filetype(self):
        """String defining the filetype."""
        return self.__class__.__name__

    def filestat(self, as_string=False):
        """
        Dictionary with file metadata
        if ``as_string`` is True, a string is returned.
        """
        d = get_filestat(self.filepath)
        if not as_string: return d
        return "\n".join("%s: %s" % (k, v) for k, v in d.items())

    @abc.abstractmethod
    def close(self):
        """Close file."""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Activated at the end of the with statement. It automatically closes the file."""
        self.close()

    #def __del__(self):
    #    """
    #    Called when the instance is about to be destroyed.
    #    """
    #    try:
    #        self.close()
    #    finally:
    #        super(BaseFile, self).__close__(self)


class TextFile(BaseFile):

    #@classmethood
    #def from_string(cls, s):
    #    return cls.from_file(filepath)

    def __enter__(self):
        # Open the file
        self._file
        return self

    def __iter__(self):
        return iter(self._file)

    @lazy_property
    def _file(self):
        """File object open in read-only mode."""
        return open(self.filepath, mode="rt")

    def close(self):
        """Close the file."""
        try:
            self._file.close()
        except Exception:
            pass

    def seek(self, offset, whence=0):
        """Set the file's current position, like stdio's fseek()."""
        self._file.seek(offset, whence)


@six.add_metaclass(abc.ABCMeta)
class AbinitNcFile(BaseFile):
    """
    Abstract class representing a Netcdf file with data saved
    according to the ETSF-IO specifications (when available).
    An AbinitNcFile has a netcdf reader to read data from file and build objects.
    """
    def ncdump(self, *nc_args, **nc_kwargs):
        """Returns a string with the output of ncdump."""
        return NcDumper(*nc_args, **nc_kwargs).dump(self.filepath)

    @lazy_property
    def abinit_version(self):
        """String with abinit version: three digits separated by comma."""
        return self.reader.rootgrp.getncattr("abinit_version")

    @abc.abstractproperty
    def params(self):
        """
        :class:`OrderedDict` with the convergence parameters
        Used to construct |pandas-DataFrames|.
        """


@six.add_metaclass(abc.ABCMeta)
class AbinitFortranFile(BaseFile):
    """
    Abstract class representing a fortran file containing output data from abinit.
    """
    def close(self):
        pass


class CubeFile(BaseFile):
    """

    .. attribute:: structure

        |Structure| object

    .. attribute:: mesh

        |Mesh3d| object with information on the uniform 3d mesh.

    .. attribute:: data

        |numpy-array| of shape [nx, ny, nz] with numerical values on the real-space mesh.
    """
    def __init__(self, filepath):
        from abipy.iotools.cube import cube_read_structure_mesh_data
        super(CubeFile, self).__init__(filepath)
        self.structure, self.mesh, self.data = cube_read_structure_mesh_data(self.filepath)

    def close(self):
        """nop, just to fulfill the abstract interface."""

    #@classmethod
    #def write_structure_mesh_data(cls, path, structure, mesh, data):
    #    with open(path, "wt") as fh:
    #        cube_write_structure_mesh(fh, structure, mesh)
    #        cube_write_data(fh, data, mesh):


@six.add_metaclass(abc.ABCMeta)
class Has_Structure(object):
    """Mixin class for :class:`AbinitNcFile` containing crystallographic data."""

    @abc.abstractproperty
    def structure(self):
        """Returns the |Structure| object."""

    def plot_bz(self, **kwargs):
        """
        Gives the plot (as a matplotlib object) of the symmetry line path in the Brillouin Zone.
        """
        return self.structure.plot_bz(**kwargs)

    # To maintain backward compatbility
    show_bz = plot_bz

    def export_structure(self, filepath):
        """
        Export the structure on file.

        returns: |Visualizer| instance.
        """
        return self.structure.export(filepath)

    def visualize_structure_with(self, appname):
        """
        Visualize the crystalline structure with the specified visualizer.

        See |Visualizer| for the list of applications and formats supported.
        """
        from abipy.iotools.visualizer import Visualizer
        visu = Visualizer.from_name(appname)

        for ext in visu.supported_extensions():
            ext = "." + ext
            try:
                return self.export_structure(ext)
            except visu.Error:
                pass
        else:
            raise visu.Error("Don't know how to export data for appname %s" % appname)

    def yield_structure_figs(self, **kwargs):
        """*Generates* a predefined list of matplotlib figures with minimal input from the user."""
        yield self.structure.plot(show=False)


@six.add_metaclass(abc.ABCMeta)
class Has_ElectronBands(object):
    """Mixin class for :class:`AbinitNcFile` containing electron data."""

    @abc.abstractproperty
    def ebands(self):
        """Returns the |ElectronBands| object."""

    @property
    def nsppol(self):
        """Number of spin polarizations"""
        return self.ebands.nsppol

    @property
    def nspinor(self):
        """Number of spinors"""
        return self.ebands.nspinor

    @property
    def nspden(self):
        """Number of indepedendent spin-density components."""
        return self.ebands.nspden

    @property
    def mband(self):
        """Maximum number of bands."""
        return self.ebands.mband

    @property
    def nband(self):
        """Maximum number of bands."""
        return self.ebands.nband

    @property
    def nelect(self):
        """Number of electrons per unit cell"""
        return self.ebands.nelect

    @property
    def nkpt(self):
        """Number of k-points."""
        return self.ebands.nkpt

    @property
    def kpoints(self):
        """Iterable with the Kpoints."""
        return self.ebands.kpoints

    @lazy_property
    def tsmear(self):
        return self.ebands.smearing.tsmear_ev.to("Ha")

    def get_ebands_params(self):
        """:class:`OrderedDict` with the convergence parameters."""
        return collections.OrderedDict([
            ("nsppol", self.nsppol),
            ("nspinor", self.nspinor),
            ("nspden", self.nspden),
            ("nband", self.nband),
            ("nkpt", self.nkpt),
        ])

    def plot_ebands(self, **kwargs):
        """Plot the electron energy bands. See the :func:`ElectronBands.plot` for the signature."""
        return self.ebands.plot(**kwargs)

    def plot_ebands_with_edos(self, edos, **kwargs):
        """Plot the electron energy bands with DOS. See the :func:`ElectronBands.plot_with_edos` for the signature."""
        return self.ebands.plot_with_edos(edos, **kwargs)

    def get_edos(self, **kwargs):
        """Compute the electronic DOS on a linear mesh. Wraps ebands.get_edos."""
        return self.ebands.get_edos(**kwargs)

    def yield_ebands_figs(self, **kwargs):
        """*Generates* a predefined list of matplotlib figures with minimal input from the user."""
        with_gaps = not self.ebands.has_metallic_scheme
        if self.ebands.kpoints.is_path:
            yield self.ebands.plot(with_gaps=with_gaps, show=False)
            yield self.ebands.kpoints.plot(show=False)
        else:
            edos = self.ebands.get_edos()
            yield self.ebands.plot_with_edos(edos, with_gaps=with_gaps, show=False)
            yield edos.plot(show=False)

    def expose_ebands(self, slide_mode=False, slide_timeout=None, **kwargs):
        """
        Shows a predefined list of matplotlib figures for electron bands with minimal input from the user.
        """
        from abipy.tools.plotting import MplExpose
        with MplExpose(slide_mode=slide_mode, slide_timeout=slide_mode, verbose=1) as e:
            e(self.yield_ebands_figs(**kwargs))


@six.add_metaclass(abc.ABCMeta)
class Has_PhononBands(object):
    """
    Mixin class for :class:`AbinitNcFile` containing phonon data.
    """

    @abc.abstractproperty
    def phbands(self):
        """Returns the |PhononBands| object."""

    def get_phbands_params(self):
        """:class:`OrderedDict` with the convergence parameters."""
        return collections.OrderedDict([
            ("nqpt", len(self.phbands.qpoints)),
        ])

    def plot_phbands(self, **kwargs):
        """
        Plot the electron energy bands. See the :func:`PhononBands.plot` for the signature.""
        """
        return self.phbands.plot(**kwargs)

    #def plot_phbands_with_phdos(self, phdos, **kwargs):
    #    return self.phbands.plot_with_phdos(phdos, **kwargs)

    def yield_phbands_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        units = kwargs.get("units", "mev")
        yield self.phbands.qpoints.plot(show=False)
        yield self.phbands.plot(units=units, show=False)
        yield self.phbands.plot_colored_matched(units=units, show=False)

    def expose_phbands(self, slide_mode=False, slide_timeout=None, **kwargs):
        """
        Shows a predefined list of matplotlib figures for phonon bands with minimal input from the user.
        """
        from abipy.tools.plotting import MplExpose
        with MplExpose(slide_mode=slide_mode, slide_timeout=slide_mode, verbose=1) as e:
            e(self.yield_phbands_figs(**kwargs))


class NcDumper(object):
    """Wrapper object for the ncdump tool."""

    def __init__(self, *nc_args, **nc_kwargs):
        """
        Args:
            nc_args: Arguments passed to ncdump.
            nc_kwargs: Keyword arguments passed to ncdump
        """
        self.nc_args = nc_args
        self.nc_kwargs = nc_kwargs
        self.ncdump = which("ncdump")

    def dump(self, filepath):
        """Returns a string with the output of ncdump."""
        if self.ncdump is None:
            return "Cannot find ncdump tool in $PATH"
        else:
            from subprocess import check_output
            return check_output(["ncdump", filepath])


_ABBREVS = [
    (1 << 50, 'Pb'),
    (1 << 40, 'Tb'),
    (1 << 30, 'Gb'),
    (1 << 20, 'Mb'),
    (1 << 10, 'kb'),
    (1, 'b'),
]


def size2str(size):
    """Convert size to string with units."""
    for factor, suffix in _ABBREVS:
        if size > factor:
            break
    return "%.2f " % (size / factor) + suffix


def get_filestat(filepath):
    stat = os.stat(filepath)
    return collections.OrderedDict([
        ("Name", os.path.basename(filepath)),
        ("Directory", os.path.dirname(filepath)),
        ("Size", size2str(stat.st_size)),
        ("Access Time", ctime(stat.st_atime)),
        ("Modification Time", ctime(stat.st_mtime)),
        ("Change Time", ctime(stat.st_ctime)),
    ])


@six.add_metaclass(abc.ABCMeta)
class NotebookWriter(object):
    """
    Mixin class for objects that are able to generate jupyter_ notebooks.
    Subclasses must provide a concrete implementation of `write_notebook`.
    """
    def make_and_open_notebook(self, nbpath=None, foreground=False):  # pragma: no cover
        """
        Generate an jupyter_ notebook and open it in the browser.

        Args:
            nbpath: If nbpath is None, a temporay file is created.
            foreground: By default, jupyter is executed in background and stdout, stderr are redirected
            to devnull. Use foreground to run the process in foreground

        Return:
            system exit code.

        Raise:
            `RuntimeError` if jupyter_ is not in $PATH
        """
        nbpath = self.write_notebook(nbpath=nbpath)

        if which("jupyter") is None:
            raise RuntimeError("Cannot find jupyter in $PATH. Install it with `conda install jupyter or `pip install jupyter`")

        # Use jupyter-lab instead of classic notebook if possible.
        has_jupyterlab = which("jupyter-lab") is not None
        #has_jupyterlab = False
        appname = "jupyter-lab" if has_jupyterlab else "jupyter notebook"

        if foreground:
            return os.system("%s %s" % (appname, nbpath))
        else:
            fd, tmpname = tempfile.mkstemp(text=True)
            print(tmpname)
            cmd = "%s %s" % (appname, nbpath)
            print("Executing:", cmd)
            print("stdout and stderr redirected to %s" % tmpname)
            import subprocess
            process = subprocess.Popen(cmd.split(), shell=False, stdout=fd, stderr=fd)
            cprint("pid: %s" % str(process.pid), "yellow")
            return 0

    @staticmethod
    def get_nbformat_nbv():
        """Return nbformat module, notebook version module"""
        import nbformat
        nbv = nbformat.v4
        return nbformat, nbv

    def get_nbformat_nbv_nb(self, title=None):
        """
        Return ``nbformat`` module, notebook version module
        and new notebook with title and import section
        """
        nbformat, nbv = self.get_nbformat_nbv()
        nb = nbv.new_notebook()

        if title is not None:
            nb.cells.append(nbv.new_markdown_cell("## %s" % title))

        nb.cells.extend([
            nbv.new_code_cell("""\
from __future__ import print_function, division, unicode_literals, absolute_import

import sys, os
import numpy as np

%matplotlib notebook
from IPython.display import display

# This to render pandas DataFrames with https://github.com/quantopian/qgrid
#import qgrid
#qgrid.nbinstall(overwrite=True)  # copies javascript dependencies to your /nbextensions folder

# This to view Mayavi visualizations. See http://docs.enthought.com/mayavi/mayavi/tips.html
#from mayavi import mlab; mlab.init_notebook(backend='x3d', width=None, height=None, local=True)

from abipy import abilab

# Tell AbiPy we are inside a notebook and use seaborn settings for plots.
# See https://seaborn.pydata.org/generated/seaborn.set.html#seaborn.set
abilab.enable_notebook(with_seaborn=True)

# AbiPy widgets for pandas and seaborn plot APIs
#import abipy.display.seabornw import snw
#import abipy.display.pandasw import pdw""")
        ])

        return nbformat, nbv, nb

    @abc.abstractmethod
    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to nbpath. If nbpath is None, a temporay file is created.
        Return path to the notebook. A typical template:

        .. code-block:: python

            # Preable.
            nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

            #####################
            # Put your code here
            nb.cells.extend([
                nbv.new_markdown_cell("# This is a markdown cell"),
                nbv.new_code_cell("a = 1"),
            ])
            #####################

            # Call _write_nb_nbpath
            return self._write_nb_nbpath(nb, nbpath)
        """

    @staticmethod
    def _write_nb_nbpath(nb, nbpath):
        """
        This method must be called at the end of ``write_notebook``.
        nb is the jupyter notebook and nbpath the argument passed to ``write_notebook``.
        """
        import io, os, tempfile
        if nbpath is None:
            _, nbpath = tempfile.mkstemp(prefix="abinb_", suffix='.ipynb', dir=os.getcwd(), text=True)

        # Write notebook
        import nbformat
        with io.open(nbpath, 'wt', encoding="utf8") as fh:
            nbformat.write(nb, fh)
            return nbpath

    @classmethod
    def pickle_load(cls, filepath):
        """
        Loads the object from a pickle file.
        """
        with open(filepath, "rb") as fh:
            new = pickle.load(fh)
            #assert cls is new.__class__
            return new

    def pickle_dump(self, filepath=None):
        """
        Save the status of the object in pickle format.
        If filepath is None, a temporary file is created.

        Return:
            name of the pickle file.
        """
        if filepath is None:
            _, filepath = tempfile.mkstemp(suffix='.pickle')

        with open(filepath, "wb") as fh:
            pickle.dump(self, fh)
            return filepath

    # TODO: Activate this
    @abc.abstractmethod
    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """

    def expose(self, slide_mode=False, slide_timeout=None, **kwargs):
        """
        Shows a predefined list of matplotlib figures with minimal input from the user.
        """
        from abipy.tools.plotting import MplExpose
        with MplExpose(slide_mode=slide_mode, slide_timeout=slide_mode, verbose=1) as e:
            e(self.yield_figs(**kwargs))


class Has_Header(object):
    """Mixin class for netcdf_ files containing the Abinit header."""

    @lazy_property
    def hdr(self):
        """|AttrDict| with the Abinit header e.g. hdr.ecut."""
        return self.reader.read_abinit_hdr()

    #def get_hdr_params(self):
    #    """:class:`OrderedDict` with the convergence parameters."""
    #    return collections.OrderedDict([

    #def compare_hdr(self, other_hdr):



#!/usr/bin/env python
r"""
DDK.nc
======

This example shows how to analyze DDK.nc files 
containing the velocity matrix elements.
"""
from abipy.abilab import abiopen
import abipy.data as abidata

# Here we use the DDK.nc files shipped with abipy.
# Replace ddk_paths with the paths to your files.
ddk_paths = abidata.ref_files(
    "gaas_444_dir1_DDK.nc",
    "gaas_444_dir2_DDK.nc",
    "gaas_444_dir3_DDK.nc",
)

from abipy.electrons.ddk import DdksAnalyzer

with DdksAnalyzer(ddk_paths) as van:
    print(van)
    #ddk_doses = van.get_doses(method="gaussian", step=0.1, width=0.2)
    #ddk_doses.plot()
    #ddk_doses.plot_with_ebands(ebands_kpath)



#!/usr/bin/env python
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import wx
import abipy.gui.awx as awx
import wx.lib.mixins.listctrl as listmix
import wx.lib.dialogs as wxdg

from monty.string import is_string
from abipy.flowtk import PseudoParser


class MyListCtrl(wx.ListCtrl, listmix.ListCtrlAutoWidthMixin):
    """ Mixin class to resize the last column appropriately."""
    def __init__(self, parent, **kwargs):
        wx.ListCtrl.__init__(self, parent, id=-1, style=wx.LC_REPORT | wx.BORDER_SUNKEN, **kwargs)
        listmix.ListCtrlAutoWidthMixin.__init__(self)


class PseudosViewerFrame(awx.Frame):
    def __init__(self, parent, pseudos, **kwargs):
        super(PseudosViewerFrame, self).__init__(parent, -1, **kwargs)

        main_sizer = wx.BoxSizer(wx.VERTICAL)

        panel = PseudoListPanel(self, pseudos)
        main_sizer.Add(panel, 1, wx.EXPAND, 5)

        #hsizer = wx.BoxSizer(wx.HORIZONTAL)

        #filter_label = wx.StaticText(self, -1, "Filter:", wx.DefaultPosition, wx.DefaultSize, 0)
        #filter_label.Wrap(-1)
        #hsizer.Add(filter_label, 0, wx.ALIGN_CENTER_VERTICAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)

        ## Combobox to enter shell patterns.
        #wildcard_choices = ["*", "*.nc", "*.abo", "*.log"]
        #self.filter_combobox = wx.ComboBox(self, id=-1, value="*", style=wx.TE_PROCESS_ENTER, choices=wildcard_choices)
        #self.filter_combobox.SetToolTipString("Shell patterns separated by |")

        #self.filter_combobox.Bind(wx.EVT_COMBOBOX, self.OnFilterComboBox)
        #self.filter_combobox.Bind(wx.EVT_TEXT_ENTER, self.OnFilterComboBox)

        #hsizer.Add(self.filter_combobox, 0, wx.ALL, 5)

        #main_sizer.Add(hsizer, 0, wx.ALIGN_CENTER_HORIZONTAL, 5)

        #panel.SetSizerAndFit(main_sizer)
        self.SetSizer(main_sizer)
        self.Layout()


class PseudoListPanel(awx.Panel, listmix.ColumnSorterMixin):
    """
    This panel shows a list of files (strings), supports column sorting
    and provides specific popup menus for the different type of files
    (file type detection is based on file extensions).
    """
    def __init__(self, parent, pseudos, **kwargs):
        """
        Args:
            parent:
                parent window
            pseudos:
                List of pseudopotentials.
        """
        super(PseudoListPanel, self).__init__(parent, -1, **kwargs)
        self.pseudos = pseudos

        self.BuildUi()

    def BuildUi(self):
        self.pseudo_list = pseudo_list = MyListCtrl(self)

        pseudo_list.Bind(wx.EVT_LIST_ITEM_RIGHT_CLICK, self.OnRightClick)
        pseudo_list.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.OnItemActivated)

        self.id2pseudo = {}
        columns = self.columns = ["basename", "symbol", "Z_val", "l_max", "rcore", "type"]

        # Used to store the Max width in pixels for the data in the column.
        column_widths = [awx.get_width_height(self, s)[0] for s in columns]

        for (index, colname) in enumerate(columns):
            pseudo_list.InsertColumn(index, colname)

        for pseudo in self.pseudos:
            entry = self.AppendPseudo(pseudo)
            w = [awx.get_width_height(self, s)[0] for s in entry]
            column_widths = map(max, zip(w, column_widths))

        for (index, colname) in enumerate(columns):
            pseudo_list.SetColumnWidth(index, column_widths[index])

        # Now that the list exists we can init the other base class, see wx/lib/mixins/listctrl.py
        self.itemDataMap = {pid: [getattr(pseudo, attr) for attr in self.columns] for pid, pseudo in self.id2pseudo.items()}
        listmix.ColumnSorterMixin.__init__(self, len(columns))
        self.Bind(wx.EVT_LIST_COL_CLICK, self.OnColClick, self.pseudo_list)

        # Pack
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(pseudo_list, 1, wx.ALL | wx.EXPAND, 5)

        self.SetSizerAndFit(sizer)

    def GetListCtrl(self):
        """Used by the ColumnSorterMixin, see wx/lib/mixins/listctrl.py"""
        return self.pseudo_list

    def AppendPseudo(self, pseudo):
        """Add a pseudo to the panel."""
        next = len(self.id2pseudo)

        # We use next as entry id because we want to be able
        # to sort the columns with the mixin ColumnSorterMixin.
        entry_id = next
        entry = map(str, [getattr(pseudo, attr) for attr in self.columns])
        self.pseudo_list.Append(entry)
        self.pseudo_list.SetItemData(next, entry_id)
        self.id2pseudo[entry_id] = pseudo

        return entry

    def OnColClick(self, event):
        event.Skip()

    def OnItemActivated(self, event):
        currentItem = event.m_itemIndex
        pseudo = self.id2pseudo[self.pseudo_list.GetItemData(currentItem)]
        print(pseudo)

        wxdg.ScrolledMessageDialog(self, str(pseudo), caption=pseudo.filepath, style=wx.MAXIMIZE_BOX).Show()

    #def OnOpenPseudoFile(self, event):
    #    """Open the pseudopotential file in another frame."""
    #    currentItem = event.m_itemIndex
    #    pseudo = self.id2pseudo[self.pseudo_list.GetItemData(currentItem)]
    #    print(pseudo)
    #    filepath = pseudo.filepath
    #    wxdg.ScrolledMessageDialog(self, str(pseudo), caption=pseudo.filepath, style=wx.MAXIMIZE_BOX).Show()

    #def OnPlotPseudo(self, event):
    #    """Open the pseudopotential file in another frame."""
    #    currentItem = event.m_itemIndex
    #    pseudo = self.id2pseudo[self.pseudo_list.GetItemData(currentItem)]
    #    print(pseudo)

    def OnRightClick(self, event):
        currentItem = event.m_itemIndex
        if currentItem == -1:
            return

        #pseudo = self.id2pseudo[self.pseudo_list.GetItemData(currentItem)]
        #print(pseudo)
        ## Open the popup menu then destroy it to avoid mem leak.
        #menu = popupmenu_for_pseudo(self, pseudo)

        #if menu is not None:
        #    self.PopupMenu(menu, event.GetPoint())
        #    menu.Destroy()


#class PopupMenu(wx.Menu):
#    """
#    Base class for popup menus. `A PopupMenu` has a list of callback functions
#    indexed by the menu title. The signature of the callback function is func(parent, filename) where
#    filename is the name of the file selected in the Widget and parent is the wx
#    Window that will become the parent of the new frame created by the callback.
#    """
#    MENU_TITLES = OrderedDict([
#    ])
#
#    HANDLED_FILES = []
#
#    def __init__(self):
#        super(PopupMenu, self).__init__()
#        self._make_menu()
#
#    @staticmethod
#    def from_filename(filename):
#        """
#        Static factory function that instanciates the appropriate subclass of `NcFilePopupMenu`
#        Returns None if the extesion of filename is not supported.
#        """
#        # Find the AbinitNcFile subclass associated to files.
#        try:
#            file_class = abifile_subclass_from_filename(filename)
#        except KeyError:
#            return None
#
#        # Check whether a subclass handles this file..
#        # Fallback to a simple PopupMenu if no match.
#        def allsubclasses(cls):
#            """Returns the set of subclasses of cls."""
#            children = [cls]
#            for sc in cls.__subclasses__():
#                if sc.__subclasses__():
#                    for k in sc.__subclasses__():
#                        children.extend(allsubclasses(k))
#                else:
#                    children.append(sc)
#            return set(children)
#
#        for cls in allsubclasses(PopupMenu):
#            if cls.handle_file_class(file_class):
#                return cls()
#        else:
#            return PopupMenu()
#
#    @classmethod
#    def handle_file_class(cls, file_class):
#        """True if the popupmenu is associated to file_class."""
#        return file_class in cls.HANDLED_FILES
#
#    def _make_menu(self):
#        """Build the menu taking into account the options of the superclasses."""
#        base_classes = list(self.__class__.__bases__) + [self.__class__]
#        base_classes.reverse()
#
#        assert not hasattr(self, "menu_title_by_id")
#        assert not hasattr(self, "menu_titles")
#        self.menu_title_by_id, self.menu_titles = OrderedDict(), OrderedDict()
#
#        for cls in base_classes:
#            try:
#                menus = cls.MENU_TITLES
#            except AttributeError as exc:
#                awx.WARNING("exc ",exc," for cls", cls)
#                continue
#
#            self.menu_titles.update(menus)
#
#            for title in menus:
#                self.menu_title_by_id[wx.NewId()] = title
#
#            # Add sentinel for Menu separator.
#            self.menu_title_by_id["separator_" + str(len(self.menu_titles))] = None
#
#        for (id, title) in self.menu_title_by_id.items():
#            if title is None:
#                sepid = int(id.split("_")[-1])
#                if sepid != len(self.menu_titles):
#                    self.AppendSeparator()
#            else:
#                # Register menu handlers with EVT_MENU, on the menu.
#                self.Append(id, title)
#                wx.EVT_MENU(self, id, self.OnMenuSelection)
#
#    def set_parent(self, parent):
#        """Set the parent window."""
#        self._parent = parent
#
#    @property
#    def parent(self):
#        """Returns the parent window."""
#        try:
#            return self._parent
#        except AttributeError:
#            awx.WARNING("Popup menu doesn't have parent")
#            return None
#
#    def set_target(self, target):
#        """Set the target of the callback."""
#        self._target = target
#
#    @property
#    def target(self):
#        """The target of the callback."""
#        try:
#            return self._target
#        except AttributeError:
#            return None
#
#    def _get_callback(self, title):
#        return self.menu_titles[title]
#
#    def OnMenuSelection(self, event):
#        title = self.menu_title_by_id[event.GetId()]
#        callback = self._get_callback(title)
#        #print("Calling callback %s on target %s" % (callback, self.target))
#        try:
#            callback(parent=self.parent, filepath=self.target)
#        except:
#            awx.showErrorMessage(parent=self.parent)
#
#
#class AbinitTextFilePopupMenu(PopupMenu):
#    """
#    """
#    MENU_TITLES = OrderedDict([
#        ("events",     showAbinitEventsFrame),
#        ("properties", showFileStat),
#        ("timer",      showAbinitTimerFrame),
#    ])
#
#    HANDLED_FILES = [AbinitLogFile, AbinitOutputFile]
#
#
#class NcFilePopupMenu(PopupMenu):
#    """
#    Base class for popup menus. `A PopupMenu` has a list of callback functions
#    indexed by the menu title and a list of `AbinitNcFile` associated to it.
#    The signature of the callback function is func(filename, parent) where
#    filename is the name of the file selected in the Widget and parent is the wx
#    Window that will become the parent of the new frame created by the callback.
#
#    How to subclass PopupMenu:
#
#        1. Define a new class that inherits from NcFilePopupMenu.
#
#        2. Define the callbacks in the class variable MENU_TITLES.
#           Use OrderedDict to have a fixed ordering of the labels.
#
#        3. Define the class variable HANDLED_FILES with the list of
#           `AbinitNcFile` subclasses associated to the popupmenu.
#
#        4. Done (most of the work is indeed done in the base class and in
#           the factory function popupmenu_for_filename.
#    """
#    MENU_TITLES = OrderedDict([
#        ("structure",  showStructure),
#        ("ncdump",     showNcdumpMessage),
#        ("properties", showFileStat),
#    ])
#
#    HANDLED_FILES = []
#
#
#class EbandsPopupMenu(NcFilePopupMenu):
#    """Popup menu for Netcdf files that contain the electron band structure."""
#    MENU_TITLES = OrderedDict([
#        ("ePlot", ewx.showElectronBandsPlot),
#        ("eDos",  ewx.showElectronDosFrame),
#        ("eJdos", ewx.showElectronJdosFrame),
#    ])
#
#    HANDLED_FILES = [WFK_File, GSR_File]


def wxapp_pseudos(dirpath):
    """Standalone application."""
    parser = PseudoParser()
    pseudos = parser.scan_directory(dirpath)

    app = wx.App()
    frame = PseudosViewerFrame(None, pseudos)
    app.SetTopWindow(frame)
    frame.Show()
    return app



#!/usr/bin/env python
"""Analyze the parallel efficiency of the BSE code (Haydock method with model dielectric function)"""
from __future__ import division, print_function, unicode_literals, absolute_import

import sys
import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_inputs(paw=False):
    # Crystalline silicon
    # Calculation of the GW correction to the direct band gap in Gamma
    # Dataset 1: ground state calculation
    # Dataset 2: BSE with haydock method and model dielectric function.
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
    pseudos = abidata.pseudos("14si.pspnc") if not paw else abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")

    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=2)

    ecut = 8
    multi.set_vars(
        ecut=ecut,
        pawecutdg=ecut*4 if paw else None,
        nsppol=1,
        timopt=-1,
        istwfk="*1",
        paral_kgb=0,
    )

    # This grid is the most economical, but does not contain the Gamma point.
    multi.set_kmesh(
        ngkpt=[8, 8, 8],
        shiftk=[0.0, 0.0, 0.0],
    )

    gs, bse = multi.split_datasets()

    gs.set_vars(tolvrs=1e-6,
                nband=8,
                )

    # Dataset 6 BSE equation with Model dielectric function and Haydock (only resonant + W + v)
    # Note that SCR file is not needed here
    bse.set_vars(
        optdriver=99,
        ecutwfn=ecut,
        ecuteps=4.0,
        inclvkb=2,
        bs_algorithm=2,        # Haydock
        bs_haydock_niter=60,   # No. of iterations for Haydock
        bs_exchange_term=1,
        bs_coulomb_term=21,    # Use model W and full W_GG.
        mdf_epsinf=12.0,
        bs_calctype=1,         # Use KS energies and orbitals to construct L0
        mbpt_sciss="0.8 eV",
        bs_coupling=0,
        bs_loband=2,
        nband=8,
        bs_freq_mesh="0 10 0.1 eV",
        bs_hayd_term=0,        # No terminator
        #gwmem=01               # Compute the model-dielectric function on-the-fly.
    )

    return gs, bse


def build_flow(options):
    """
    Build an `AbinitWorkflow` used for benchmarking ABINIT.
    """
    gs_inp, bse_inp = make_inputs(paw=options.paw)
    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)

    gs_work = flowtk.Work()
    gs_work.register_scf_task(gs_inp)
    flow.register_work(gs_work)
    flow.exclude_from_benchmark(gs_work)

    mpi_list = options.mpi_list

    if options.mpi_list is None:
        nkpt = len(gs_inp.abiget_ibz().points)
        ntrans = (2*2*nkpt)**2
        mpi_list = [p for p in range(1, 1 + ntrans) if ntrans % p == 0]
    if options.verbose: print("Using mpi_list:", mpi_list)

    bse_work = flowtk.Work()
    for mpi_procs, omp_threads in product(mpi_list, options.omp_list):
        if not options.accept_mpi_omp(mpi_procs, omp_threads): continue
        manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
        bse_work.register_bse_task(bse_inp, manager=manager, deps={gs_work[0]: "WFK"})
    flow.register_work(bse_work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



from __future__ import print_function, division, unicode_literals, absolute_import

import wx
import abipy.gui.awx as awx

from pymatgen.analysis.eos import EOS


class EosFrame(awx.Frame):
    """
    Frame that allows the user to fit E(V).
    """
    def __init__(self, parent, volumes, energies, vol_unit="ang^3", ene_unit="eV", **kwargs):
        super(EosFrame, self).__init__(parent, id=-1, title="EOS Frame", **kwargs)

        self.volumes, self.vol_unit = volumes, vol_unit
        self.energies, self.ene_unit = energies, ene_unit

        self.BuildUi()

    def BuildUi(self):
        main_sizer = wx.BoxSizer(wx.VERTICAL)

        models = EOS.MODELS.keys()
        self.model_choice = wx.Choice(self, -1, wx.DefaultPosition, wx.DefaultSize, models, 0)
        self.model_choice.SetSelection(0)
        main_sizer.Add(self.model_choice, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL | wx.EXPAND, 5)

        fit_button = wx.Button(self, -1, "Plot", wx.DefaultPosition, wx.DefaultSize, 0)
        fit_button.Bind(wx.EVT_BUTTON, self.OnFitButton)
        main_sizer.Add(fit_button, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)

        self.SetSizerAndFit(main_sizer)

    def OnFitButton(self, event):
        model = self.model_choice.GetStringSelection()

        try:
            eos = EOS(eos_name=model)
            fit = eos.fit(self.volumes, self.energies, vol_unit=self.vol_unit, ene_unit=self.ene_unit)
            print(fit)
            fit.plot()

        except:
            awx.showErrorMessage(self)




# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

executable = "optic"

from abipy.abio.abivar_database.variables import ValueWithUnit, MultipleValue, Range, ValueWithConditions
ValueWithConditions = dict

Variable=dict
variables = [
Variable(
    abivarname="broadening@optic",
    varset="optic",
    vartype="real",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval="1.d-3 Ha",
    mnemonics="BROADENING",
    text="""
In Eq. 46 of [[cite:Sharma2004], it is clear that when ever wnm(k) is equal to w, there is a resonance.
Numerically this would lead to an infinity. In order to avoid this one could do two things.

You could change the sum over k-points to integration and then use
linear tetrahedron method (see [[cite:Hughes1996]] for details).
Another way to get around the problem is, like we do in the present case, avoid this
singularity by adding a small complex number to the denominator. This prevents
the denominator from ever going to 0 and acts as a broadening to the spectrum.
The broadening should not be too large as this would wash out the features in the spectrum.
""",
),

Variable(
    abivarname="ddkfile@optic",
    varset="optic",
    vartype="string",
    topics=['Optic_basic'],
    dimensions="scalar",
    mnemonics="DDK FILE",
    commentdefault="no default",
    text="""
Specify the filename that has been produced by the preparatory Abinit run.
This file must contain the matrix elements of the d/dk operator along direction X.
It must not contain the first-order wavefunctions and may be generated using [[prtwf]] 3.
You should make sure that the number of bands, of spin channels and of
k-points are the same in all the files.

use as string with the filename: ddkfile_X, where X is the file number.
""",
),

Variable(
    abivarname="domega@optic",
    varset="optic",
    vartype="real",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval="1.d-3 Ha",
    mnemonics="Delta OMEGA",
    text="""
The step and maximum sets your energy grid for the calculation using the
formula number of energy mesh points=maximum/step (zero excluded). So in order
to capture more features you can decrease the step size to get a finer energy
grid. In order to go to higher frequency, increase the maximum.
""",
),

Variable(
    abivarname="lin_comp@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions=[['num_lin_comp']],
    defaultval=0,
    mnemonics="LINear COMPonents",
    text="""
This tells which component of the dielectric tensor you want to calculate.
These numbers are called a and b Eqs. 46 in [[cite:Sharma2004]].
1 2 3 represent x y and z respectively. For example 11 would be xx and 32 would mean zy.
""",
),

Variable(
    abivarname="maxomega@optic",
    varset="optic",
    vartype="real",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval="1 Ha",
    mnemonics="MAXimum value of OMEGA",
    text="""
The step and maximum sets your energy grid for the calculation using the
formula number of energy mesh points=maximum/step (zero excluded). So in order
to capture more features you can decrease the step size to get a finer energy grid.
In order to go to higher frequency, increase the maximum.
""",
),

Variable(
    abivarname="nonlin_comp@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions=[['num_nonlin_comp']],
    defaultval=0,
    mnemonics="NON-LINear COMPonents",
    text="""
This tells which component of the dielectric tensor you want to calculate.
These numbers are called a, b and c in [[cite:Sharma2004]].
1 2 3 represent x y and z respectively. For example 111 would be xxx and 321 would mean zyx.
""",
),

Variable(
    abivarname="num_lin_comp@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval=0,
    mnemonics="NUMber of LINear COMPonents",
    text="""
How many components out of 9 of the linear optical dielectric tensor do you
want to calculate. Most of these are either equal or zero depending upon the
symmetry of the material (for detail see [[cite:Draxl2006]]).
Note that the directions are along the Cartesian axis.
""",
),

Variable(
    abivarname="num_nonlin_comp@optic",
    varset="optic",
    vartype="integer",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval=0,
    mnemonics="NUMber of NON-LINear COMPonents",
    text="""
How many components out of 27 of the non-linear optical dielectric tensor do
you want to calculate. Most of these are either equal or zero depending upon
the symmetry of the material (for detail see [[cite:Draxl2006]]).
Note that the directions are along the Cartesian axis.
""",
),

Variable(
    abivarname="scissor@optic",
    varset="optic",
    vartype="real",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval=0.0,
    mnemonics="SCISSOR operator",
    commentdefault="in Ha",
    text="""
LDA/GGA are known to underestimate the band-gap by up to 100%. In order
to get the optical spectrum and make a realistic comparison with experiments
one needs to correct for this. This can be achieved in two ways.

The scissors shift is normally chosen to be the difference between the experimental and
theoretical band-gap and is used to shift the conduction bands only. Another
way in which you do not have to rely on experimental data is to determine the
self energy using the [[lesson:gw1|GW approach]].
In this case the opening of the gap due to the GW correction can be used as scissor shift.
""",
),

Variable(
    abivarname="tolerance@optic",
    varset="optic",
    vartype="real",
    topics=['Optic_basic'],
    dimensions="scalar",
    defaultval="1.d-3 Ha",
    mnemonics="TOLERANCE",
    text="""
When energy denominators are smaller than **tolerance** , the term is discarded from the sum.
""",
),

Variable(
    abivarname="wfkfile@optic",
    varset="optic",
    vartype="string",
    topics=['Optic_basic'],
    dimensions="scalar",
    mnemonics="WaveFunction K FILE",
    commentdefault="no default",
    text="""
Specify the filename that has been produced by the preparatory Abinit run.
This file must contain the matrix elements of the d/dk operator along
direction X. It must not contain the first-order wavefunctions and may be
generated using [[prtwf]] 3.
You should make sure that the number of bands, of spin channels and of
k-points are the same in all the files.
""",
),

]



from __future__ import print_function, division, unicode_literals, absolute_import

import wx
import pymatgen.core.units as units

from abipy.gui import awx


class ConverterFrame(awx.Frame):
    DEFAULT_UTYPE = "energy"

    def __init__(self, parent, **kwargs):
        super(ConverterFrame, self).__init__(parent, -1, title="Unit Converter")
        self.BuildUi()

    def BuildUi(self):
        main_sizer = wx.BoxSizer(wx.VERTICAL)

        hsz1 = wx.BoxSizer(wx.HORIZONTAL)

        from_text = wx.StaticText(self, -1, "From:")
        from_text.Wrap(-1)
        hsz1.Add(from_text, 0, wx.ALL, 5)

        self.from_textctrl = wx.TextCtrl(self, -1, "1.0")
        hsz1.Add(self.from_textctrl, 1, wx.ALL, 5)

        from_uchoices = units.ALL_UNITS[self.DEFAULT_UTYPE].keys()
        self.from_unit_choice = wx.ComboBox(self, -1, choices=from_uchoices)
        self.from_unit_choice.SetSelection(0)
        hsz1.Add(self.from_unit_choice, 1, wx.ALL, 5)

        main_sizer.Add(hsz1, 0.4, wx.ALIGN_CENTER_HORIZONTAL | wx.EXPAND, 5)

        hsz2 = wx.BoxSizer(wx.HORIZONTAL)

        to_text = wx.StaticText(self, -1, "To:")
        to_text.Wrap(-1)
        hsz2.Add(to_text, 0, wx.ALL, 5)

        self.to_textctrl = wx.TextCtrl(self, -1, wx.EmptyString)
        hsz2.Add(self.to_textctrl, 1, wx.ALL, 5)

        to_uchoices = units.ALL_UNITS[self.DEFAULT_UTYPE].keys()
        self.to_unit_choice = wx.ComboBox(self, -1, choices=to_uchoices)
        self.to_unit_choice.SetSelection(0)
        hsz2.Add(self.to_unit_choice, 1, wx.ALL, 5)

        main_sizer.Add(hsz2, 0, wx.ALIGN_CENTER_HORIZONTAL | wx.EXPAND, 5)

        hsz3 = wx.BoxSizer(wx.HORIZONTAL)

        utype_text = wx.StaticText(self, -1, "Unit type:")
        utype_text.Wrap(-1)
        hsz3.Add(utype_text, 0, wx.ALL, 5)

        unit_type_choices = units.ALL_UNITS.keys()
        self.unit_type_choice = wx.Choice(self, -1, wx.DefaultPosition, wx.DefaultSize, unit_type_choices, 0)
        self.unit_type_choice.SetSelection(unit_type_choices.index(self.DEFAULT_UTYPE))
        self.Bind(wx.EVT_CHOICE, self.OnUnitTypeChoiced, self.unit_type_choice)

        hsz3.Add(self.unit_type_choice, 0, wx.ALL, 5)

        convert_button = wx.Button(self, -1, "Convert")
        hsz3.Add(convert_button, 0, wx.ALL, 5)

        main_sizer.Add(hsz3, 1, wx.EXPAND, 5)

        self.Bind(wx.EVT_BUTTON, self.OnConvert, convert_button)

        self.SetSizerAndFit(main_sizer)

    def OnUnitTypeChoiced(self, event):
        new_type = self.unit_type_choice.GetStringSelection()
        new_choices = units.ALL_UNITS[new_type].keys()

        self.from_unit_choice.Clear()
        self.from_unit_choice.AppendItems(new_choices)
        self.from_unit_choice.SetSelection(0)

        self.to_unit_choice.Clear()
        self.to_unit_choice.AppendItems(new_choices)
        self.to_unit_choice.SetSelection(0)

    def OnConvert(self, event):
        try:
            fvalue = float(self.from_textctrl.GetValue())
        except:
            awx.showErrorMessage(self)
            return

        from_unit = self.from_unit_choice.GetValue()
        if from_unit == "foo":
            from abipy.gui.awx.eggs import Puzzle
            return Puzzle(self)

        ufloat = units.FloatWithUnit(fvalue, from_unit)
        to_unit = self.to_unit_choice.GetValue()

        try:
            conversion = ufloat.to(to_unit)
            print(ufloat, conversion)
            self.to_textctrl.SetValue(str(float(conversion)))

        except:
            awx.showErrorMessage(self)


def wxapp_converter():
    """Standalon application."""
    app = awx.App()
    frame = ConverterFrame(None)
    frame.Show()
    app.SetTopWindow(frame)
    return app


if __name__ == "__main__":
    wxapp_converter().MainLoop()



#!/usr/bin/env python
r"""
Flow for quasi-harmonic calculations
====================================

Warning: This code is still under development.
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata
from abipy import flowtk

from abipy.flowtk.qha import QhaFlow


def build_flow(options):
    """
    Create a `QhaFlow` for quasi-harmonic calculations.
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    # Initialize structure and pseudos
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
    pseudos = abidata.pseudos("14si.pspnc")

    # Build input for GS calculation.
    gsinp = abilab.AbinitInput(structure, pseudos)
    gsinp.set_vars(ecut=4, nband=4, toldff=1.e-6)
    gsinp.set_autokmesh(nksmall=2)

    volumes = [gsinp.structure.volume]
    return QhaFlow.from_gsinp(options.workdir, gsinp, volumes, ngqpt=[2,2,2])


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("GENERATE_SPHINX_GALLERY", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).plot_networkx(tight_layout=True)


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



from __future__ import print_function, division, unicode_literals, absolute_import

import os
import wx
import abc
import abipy.gui.awx as awx

from wx.py.shell import Shell
from monty.string import marquee
from abipy.abilab import abiopen
from abipy.gui import mixins as mix
from abipy.gui.kpoints import KpointsPanel
from abipy.gui.baseviewer import MultiViewerFrame


class PhbstViewerFrame(MultiViewerFrame, mix.Has_Structure, mix.Has_MultiplePhbands, mix.Has_Tools, mix.Has_NetcdfFiles):
    VERSION = "0.1"

    HELP_MSG = """Quick help:

 Kpoint list:

     Right-Click:  display popup menu with choices.

Also, these key bindings can be used
(For Mac OSX, replace 'Ctrl' with 'Apple'):

  Ctrl-Q:     quit
"""

    @property
    def codename(self):
        return "PhbstViewer"

    @property
    def active_phbst(self):
        """The active PHBST file i.e. the PHBST associated to the active tab."""
        return self.active_tab.phbst

    @property
    def structure(self):
        """`Structure` associated to the active tab."""
        return self.active_phbst.structure

    @property
    def phbands(self):
        """`PhononBands` associated to the active tab."""
        return self.active_phbst.phbands

    @property
    def active_phdos_file(self):
        """The active PHDOS file i.e. the PHDOS associated to the active tab."""
        try:
            return self.active_tab.phdos_file
        except AttributeError:
            return None

    @property
    def phdos(self):
        """PHDOS data for the active tab if it has been added. None otherwise"""
        if self.active_phdos_file:
            return self.active_phdos_file.phdos
        else:
            return None

    @property
    def phbands_list(self):
        """List of `PhononBands`."""
        phbands_list = []
        for page in range(self.notebook.GetPageCount()):
            tab = self.notebook.GetPage(page)
            phbands_list.append(tab.phbst.phbands)
        return phbands_list

    @property
    def phbands_filepaths(self):
        """
        Return a list with the absolute paths of the files
        from which the `PhononBands` have been read.
        """
        paths = []
        for page in range(self.notebook.GetPageCount()):
            tab = self.notebook.GetPage(page)
            paths.append(tab.phbst.filepath)
        return paths

    @property
    def phdos_list(self):
        """List of `PhononDos`."""
        phdos_list = []
        for page in range(self.notebook.GetPageCount()):
            tab = self.notebook.GetPage(page)
            if tab.phdos_file:
                phdos_list.append(tab.phdos_file.phdos)
        return phdos_list

    @property
    def phdos_filepaths(self):
        """
        Return a list with the absolute paths of the files
        from which the `PhononDos` have been read.
        """
        paths = []
        for page in range(self.notebook.GetPageCount()):
            tab = self.notebook.GetPage(page)
            if tab.phdos_file:
                paths.append(tab.phdos_file.filepath)
        return paths

    @property
    def nc_filepaths(self):
        """String with the absolute path of the netcdf file."""
        paths = []
        for page in range(self.notebook.GetPageCount()):
            tab = self.notebook.GetPage(page)
            paths.append(tab.phbst.filepath)
        return paths

    def makeMenu(self):
        """Creates the main menu."""
        # Base menu
        menu_bar = super(PhbstViewerFrame, self).makeMenu()

        # Add Mixin menus.
        menu_bar.Append(self.CreateStructureMenu(), "Structure")
        menu_bar.Append(self.CreatePhbandsMenu(), "Phbands")
        menu_bar.Append(self.CreateToolsMenu(), "Tools")
        menu_bar.Append(self.CreateNetcdfMenu(), "Netcdf")

        # Help menu
        help_menu = self.makeHelpMenu()
        menu_bar.Append(help_menu, "Help")

        self.SetMenuBar(menu_bar)

    def makeToolBar(self):
        """Creates the toolbar."""
        self.toolbar = toolbar = self.CreateToolBar()
        toolbar.SetToolBitmapSize(wx.Size(48, 48))

        def bitmap(path):
            return wx.Bitmap(awx.path_img(path))

        artBmp = wx.ArtProvider.GetBitmap
        toolbar.AddSimpleTool(wx.ID_OPEN, artBmp(wx.ART_FILE_OPEN, wx.ART_TOOLBAR), "Open")
        #toolbar.AddSeparator()
        toolbar.Realize()

    def addFileTab(self, parent, filepath):
        """Read data from filepath and create a new notebook tab."""
        phbst = abiopen(filepath)
        tab = PhbstFileTab(self.notebook, phbst)
        self.notebook.AddPage(tab, os.path.basename(filepath))


class PhbstFileTab(wx.Panel):
    """Tab showing information on a single PHBST file."""
    def __init__(self, parent, phbst, **kwargs):
        """
        Args:
            parent:
                parent window.
            phbst:
                `PhbstFile` instance.
        """
        super(PhbstFileTab, self).__init__(parent, -1, **kwargs)
        self.phbst = phbst
        self.phdos_file = None

        splitter = wx.SplitterWindow(self, id=-1, style=wx.SP_3D)
        splitter.SetSashGravity(0.95)

        self.qpoints_panel = KpointsPanel(splitter, phbst.structure, phbst.qpoints)

        # Add Python shell
        msg = "PHBST_File object is accessible via the phbst variable. Use phbst.<TAB> to access the list of methods."
        msg = marquee(msg, width=len(msg) + 8, mark="#")
        msg = "#"*len(msg) + "\n" + msg + "\n" + "#"*len(msg) + "\n"

        pyshell = Shell(splitter, introText=msg, locals={"phbst": self.phbst})
        splitter.SplitHorizontally(self.qpoints_panel, pyshell)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(splitter, 1, wx.EXPAND, 5)
        self.SetSizerAndFit(sizer)

    @property
    def statusbar(self):
        return self.viewer_frame.statusbar

    def GetVisualizer(self):
        """Returns a string with the visualizer selected by the user."""
        return self.viewer_frame.GetVisualizer()

    @property
    def viewer_frame(self):
        """The parent frame `PhbstViewerFrame`."""
        try:
            return self._viewer_frame

        except AttributeError:
            self._viewer_frame = self.getParentWithType(PhbstViewerFrame)
            return self._viewer_frame


class PhbstViewerApp(awx.App):
    pass


def wxapp_phbstviewer(phbst_filepaths):
    """Standalone application."""
    app = PhbstViewerApp()
    if not phbst_filepaths:
        phbst_filepaths = ()
    frame = PhbstViewerFrame(None, filepaths=phbst_filepaths)
    app.SetTopWindow(frame)
    frame.Show()
    return app



# coding: utf-8
"""
Tools and workflows for calculations within the quasi-harmonic approximation.

WARNING: This code is still under development.
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import numpy as np

from pymatgen.io.abinit.flows import Flow
from pymatgen.io.abinit.works import Work, RelaxWork, PhononWork, MergeDdb
from abipy.core.structure import Structure
from abipy.abio.inputs import AbinitInput

import logging
logger = logging.getLogger(__name__)


class RelaxAndPhononWork(Work, MergeDdb):
    """
    This work perform a structural relaxation and use the relaxed structure to
    compute phonon frequencies with DFPT.
    """
    _relax_done, _phonons_done = False, False

    @classmethod
    def from_gsinp(cls, gsinp, ngqpt, optcell, ionmov):
        work = cls()
        work.initial_gsinp, work.ngqpt = gsinp, ngqpt
        # Create input for relaxation and register relaxation task.
        relaxinp = gsinp.deepcopy()
        relaxinp.pop_tolerances()
        relaxinp.set_vars(tolvrs=1e-10, toldff=1.e-6)
        relaxinp.set_vars(optcell=optcell, ionmov=ionmov)
        if optcell is not None and optcell != 0 :
            relaxinp.set_vars_ifnotin(ecutsm=0.5, dilatmx=1.05)

        work.relax_task = work.register_relax_task(relaxinp)
        return work

    def add_phonon_tasks_and_build(self):
        """
        Called by on_all_ok when the structural relaxation is completed.
        Here the phonon tasks with the relaxed structure and constructed
        and added to the work.
        """
        # Get relaxed structure and build tasks for phonon calculations.
        relaxed_structure = self.relax_task.get_final_structure()
        scf_input = self.initial_gsinp.new_with_structure(relaxed_structure)
        scf_task = self.register_scf_task(scf_input)

        qpoints = scf_input.abiget_ibz(ngkpt=self.ngqpt, shiftk=[0, 0, 0]).points
        for qpt in qpoints:
            for ph_inp in scf_task.input.make_ph_inputs_qpoint(qpt, tolerance=None):
                self.register_phonon_task(ph_inp, deps={scf_task: "WFK"})

        # Allocate new works and update the pickle database.
        self.flow.allocate()
        self.flow.build_and_pickle_dump()

    def on_all_ok(self):
        """
        This method is called when all the task reach S_OK.
        The first time we enter here, we construct the phonon tasks.
        The second call runs `mrgddb` in sequential on the local machine to produce
        the final DDB file in the outdir of the `Work`.
        """
        if not self._relax_done:
            # Build phonon tasks.
            self._relax_done = True
            self.add_phonon_tasks_and_build()
            self.finalized = False

        elif not self._phonons_done:
            # Merge DDB files.
            self._phonons_done = True
            self.merge_ddb_files()
            self.finalized = True

        return super(RelaxAndPhononWork, self).on_all_ok()


class QhaFlow(Flow):

    @classmethod
    def from_gsinp(cls, workdir, gsinp, volumes, ngqpt, manager=None):
        """

        Args:
            workdir:
            gsinp:
            volumes:
            ngqpt:
            manager:
        """
        ngqpt = np.reshape(ngqpt, 3)
        flow = cls(workdir=workdir, manager=manager)

        # Construct len(volumes) works. Each work performs the structure relaxation
        # at fixed volume followed by DFPT calculation with the relaxed structure.
        for vol in volumes:
            # Build GS input file for new structure with rescaled volume.
            new_lattice = gsinp.structure.lattice.scale(vol)
            new_structure = Structure(new_lattice, gsinp.structure.species, gsinp.structure.frac_coords)
            new_input = gsinp.new_with_structure(new_structure)

            # Register work.
            work = RelaxAndPhononWork.from_gsinp(new_input, ngqpt, optcell=3, ionmov=3)
            flow.register_work(work)

        return flow

    def finalize(self):
        """
        This method is called when the flow is completed.
        It transfer the most important results to flow/outdir and perform
        some basic post-processing of the results to facilitate further analysis.
        """
        #data = {}
        #for work in self:
        #    ddb_path = work.outdir.has_abiext("DDB")
        #    with work.relax_task.open_gsr() as gsr
        #        data.append(dict(
        #            volume=gsr.structure.volume,
        #            etotal=gsr.etotal,
        #            structure=gsr.structure,
        #            pressure_GPa=gsr.pressure,
        #            stress_tensor=gsr.stress_tensor,
        #            forces=gsr.forces,
        #            ddb_path=ddb_path,
        #            gsr_path=gsr.path,
        #        ))

        #with open(self.outdir.path_in("qha.sjon"), "wt") as fh:
        #    json.dump(fh, data)

        return super(QhaFlow, self).finalize()



"""
This module provides functions to generate crystalline structures from Wyckoff positions
or to retrieve Wyckoff parameters from a given structure.
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import numpy as np

#from monty.functools import lazy_property
from collections import Sequence, OrderedDict


def _validate_params(params, wpos):
    """
    Check keys in params. Raises ValueError.
    """
    check = params.copy()
    for k in params:
        if k not in wpos.names:
            raise ValueError("Cannot find key: %s in symbol_names: %s" % (k, wpos.names))
        check.pop(k)
    if check:
        raise ValueError("Found unknown symbol names in %s" % check)


class WyckoffPositions(Sequence):
    """
    A numpy array with sympy expression.
    """
    @classmethod
    def from_string(cls, s):
        """
        Initialize the object from string `s`.
        """
        # Build a matrix of strings with expressions.
        s = s.replace("\n", "")
        s = s.replace("(", "").replace(")", "")
        s = s.replace("[", "").replace("]", "")

        try:
            matrix_str = np.reshape(s.split(","), (-1, 3))
        except ValueError:
            raise ValueError("Wrong string in input, perhaps missing comma. Input string:\n %s" % str(s))

        return cls(matrix_str)

    def __init__(self, matrix_str, letter=None, site_symmetry=None):
        """

        Args:
            matrix_str:
            letter:
            site_symmetry:
        """
        self.letter, self.site_symmetry = None, None

        # Build numpy matrix of sympy expressions from the matrix of strings.
        from sympy import Symbol
        from sympy.parsing.sympy_parser import parse_expr
        exprs = []
        for vec in np.reshape(matrix_str, (-1, 3)):
            exprs.extend([parse_expr(e) for e in vec])

        self._expr_mat = np.reshape(exprs, (-1, 3))

        # Get set of sympy symbols.
        symbols = []
        for expr in self.ravel():
            for arg in expr.args:
                if isinstance(arg, Symbol): symbols.append(arg)

        # SymPy symbols and names.
        self.symbols = list(set(symbols))
        self.names = [s.name for s in self.symbols]

    def __len__(self):
        return len(self._expr_mat)

    def __getitem__(self, key):
        return self._expr_mat.__getitem__(key)

    def __str__(self):
        """String representation."""
        lines = []
        for vec in self:
            lines.append("(" + ",".join(str(e) for e in vec) + ")")
        return ", ".join(lines)

    def ravel(self):
        """Flat list, similar to np.ravel"""
        return self._expr_mat.ravel()

    @property
    def mult(self):
        """Multiplicity"""
        return len(self)

    @property
    def num_params(self):
        """The number of parameters."""
        return len(self.symbols)

    def params_from_frac_coords(self, frac_coords):
        """
        Compute the Wyckoff parameters from the fractional coordinates.
        """
        frac_coords = np.reshape(frac_coords, (-1, 3))
        assert len(self) == len(frac_coords)
        equations = []
        for i in range(len(self)):
            eqs = self[i, :] - frac_coords[i, :]
            equations.extend(eqs)

        # Solve the problem.
        from sympy.solvers import solve
        d = solve(equations, self.symbols)
        #print("solution: ", d)

        # Return results in a dict with symbol names.
        if not d:
            return None

        return {symbol.name: d[symbol] for symbol in self.symbols}

    def error_of_params(self, params, frac_coords):
        """
        Return a numpy array with the difference between the
        wyckoff positions computed from the sympy expressions and the input frac_coords
        """
        _validate_params(params, self)
        frac_coords = np.reshape(frac_coords, (-1, 3))
        assert len(self) == len(frac_coords)

        frac_errors = np.empty((self.mult, 3))
        for i in range(len(self)):
            for j in range(3):
                frac_errors[i,j] = self[i,j].subs(params) - frac_coords[i,j]

        # Modulo lattice vector (with tolerance?)
        return frac_errors % 1


class Wyckoff(object):
    """
    """
    #@classmethod
    #def from_site2wpos_ordict(cls, site2wpos):
    #    return cls(site2wpos)

    def __init__(self, site2wpos):
        # TODO: Ordered dict.
        d = {}
        for k, s in site2wpos.items():
            d[k] = WyckoffPositions.from_string(s)

        self.site2wpos = d

    def __str__(self):
        """String representation."""
        lines = []
        app = lines.append
        app("Wyckoff positions:")
        for site, wpos in self.site2wpos.items():
            app("%s site:\n\t%s" % (site, wpos))

        return "\n".join(lines)

    def generate_structure(self, lattice, site2params, to_unit_cell=False, struct_cls=None):
        """
        Generate structure object from `lattice` and dictionary `site2params`
        """
        if not isinstance(site2params, OrderedDict):
            raise ValueError("Please use a OrderedDict for site2params so that\n" +
                             "the algorithm used to build the structure is deterministic.")

        # Build species and coords arrays.
        species, coords = [], []
        for elsym, wpos in self.site2wpos.items():
            params = site2params[elsym]
            _validate_params(params, wpos)
            for vec_expr in wpos:
                species.append(elsym)
                # Evaluate sympy expression with params.
                coords.append([float(expr.subs(params)) for expr in vec_expr])

        #for sp, c in zip(species, coords):
        #    print(sp, c, type(c), type(c[0]))

        from abipy.core.structure import Structure
        cls = struct_cls if struct_cls is not None else Structure
        return cls(lattice, species, coords, validate_proximity=True,
                   to_unit_cell=to_unit_cell, coords_are_cartesian=False, site_properties=None)

    def find_params(self, structure):
        """
        Compute the value of the parameter given a Structure object.
        """
        frac_coords = structure.get_symbol2coords()

        # Solve the problem.
        params = {}
        for elsym, wpos in self.site2wpos.items():
            params[elsym] = wpos.params_from_frac_coords(frac_coords[elsym])

        # Return results in a dict with symbol names.
        return params

    def error_of_params(self, params, structure):
        frac_coords = structure.get_symbol2coords()

        for elsym, wpos in self.site2wpos.items():
            frac_errors = wpos.error_of_params(params[elsym], frac_coords[elsym])
            print(frac_errors)



#!/usr/bin/env python
r"""
G0W0 Flow with convergence study wrt nband
==========================================

This script shows how to compute the G0W0 corrections in silicon.
More specifically, we build a flow to analyze the convergence of the QP corrections
wrt to the number of bands in the self-energy. More complicated convergence studies
can be implemented on the basis of this example.
"""
from __future__ import division, print_function, unicode_literals, absolute_import

import os
import sys
import abipy.data as data
import abipy.abilab as abilab
from abipy import flowtk


def make_inputs(ngkpt, paral_kgb=1):
    # Crystalline silicon
    # Calculation of the GW correction to the direct band gap in Gamma
    # Dataset 1: ground state calculation
    # Dataset 2: NSCF calculation
    # Dataset 3: calculation of the screening
    # Dataset 4-5-6: Self-Energy matrix elements (GW corrections) with different values of nband

    multi = abilab.MultiDataset(structure=data.cif_file("si.cif"), pseudos=data.pseudos("14si.pspnc"), ndtset=6)

    # This grid is the most economical, but does not contain the Gamma point.
    scf_kmesh = dict(
        ngkpt=ngkpt,
        shiftk=[0.5, 0.5, 0.5,
                0.5, 0.0, 0.0,
                0.0, 0.5, 0.0,
                0.0, 0.0, 0.5]
    )

    # This grid contains the Gamma point, which is the point at which
    # we will compute the (direct) band gap.
    gw_kmesh = dict(
        ngkpt=ngkpt,
        shiftk=[0.0, 0.0, 0.0,
                0.0, 0.5, 0.5,
                0.5, 0.0, 0.5,
                0.5, 0.5, 0.0]
    )

    # Global variables. gw_kmesh is used in all datasets except DATASET 1.
    ecut = 6

    multi.set_vars(
        ecut=ecut,
        timopt=-1,
        istwfk="*1",
        paral_kgb=paral_kgb,
        gwpara=2,
    )
    multi.set_kmesh(**gw_kmesh)

    # Dataset 1 (GS run)
    multi[0].set_kmesh(**scf_kmesh)
    multi[0].set_vars(
        tolvrs=1e-6,
        nband=4,
    )

    # Dataset 2 (NSCF run)
    # Here we select the second dataset directly with the syntax multi[1]
    multi[1].set_vars(
        iscf=-2,
        tolwfr=1e-12,
        nband=35,
        nbdbuf=5,
    )

    # Dataset3: Calculation of the screening.
    multi[2].set_vars(
        optdriver=3,
        nband=25,
        ecutwfn=ecut,
        symchi=1,
        inclvkb=0,
        ecuteps=4.0,
        ppmfrq="16.7 eV",
    )

    # Dataset4: Calculation of the Self-Energy matrix elements (GW corrections)
    kptgw = [
         -2.50000000E-01, -2.50000000E-01,  0.00000000E+00,
         -2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
          5.00000000E-01,  5.00000000E-01,  0.00000000E+00,
         -2.50000000E-01,  5.00000000E-01,  2.50000000E-01,
          5.00000000E-01,  0.00000000E+00,  0.00000000E+00,
          0.00000000E+00,  0.00000000E+00,  0.00000000E+00,
      ]

    bdgw = [1, 8]

    # Convergence study wrt nband in sigma.
    for idx, nband in enumerate([10, 20, 30]):
        multi[3+idx].set_vars(
            optdriver=4,
            nband=nband,
            ecutwfn=ecut,
            ecuteps=4.0,
            ecutsigx=6.0,
            symsigma=1,
            #gw_qprange=0,
            #nkptgw=0,
        )
        multi[3+idx].set_kptgw(kptgw, bdgw)

    return multi.split_datasets()


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_","flow_")

    # Change the value of ngkpt below to perform a GW calculation with a different k-mesh.
    scf, nscf, scr, sig1, sig2, sig3 = make_inputs(ngkpt=[2, 2, 2])

    return flowtk.g0w0_flow(options.workdir, scf, nscf, scr, [sig1, sig2, sig3], manager=options.manager)


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    #build_flow(options).plot_networkx(with_edge_labels=True, tight_layout=True)
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_si_g0w0.py -s
#
# The last three tasks (``w0_t3``, ``w0_t4``, ``w0_t5``) are the SigmaTask who have produced
# a netcdf file with the GW results with different number of bands.
# We can check this with the command:
#
#    abirun.py flow_si_g0w0/ listext SIGRES
#
# .. code-block:: bash
#
#       Found 3 files with extension `SIGRES` produced by the flow
#       File                                        Size [Mb]    Node_ID  Node Class
#       ----------------------------------------  -----------  ---------  ------------
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc         0.05     241325  SigmaTask
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc         0.08     241326  SigmaTask
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc         0.13     241327  SigmaTask
#
# Let's use the SIGRES robot to collect and analyze the results:
#
#    abirun.py flow_si_g0w0/ robot SIGRES
#
# and then, inside the ipython terminal, type:
#
# .. code-block:: ipython
#
#       In [1]: df = robot.get_dataframe()
#       In [2]: df
#       Out[2]:
#                                                 nsppol     qpgap            ecutwfn  \
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc       1  3.627960  5.914381651684836
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc       1  3.531781  5.914381651684836
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc       1  3.512285  5.914381651684836
#
#                                                            ecuteps  \
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc  3.6964885323070074
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc  3.6964885323070074
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc  3.6964885323070074
#
#                                                          ecutsigx scr_nband  \
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc  5.914381651684846        25
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc  5.914381651684846        25
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc  5.914381651684846        25
#
#                                                sigma_nband gwcalctyp scissor_ene  \
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc          10         0         0.0
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc          20         0         0.0
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc          30         0         0.0
#
#                                                 nkibz
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc      6
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc      6
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc      6
#
#       In [3]: %matplotlib
#       In [4]: df.plot("sigma_nband", "qpgap", marker="o")
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_si_g0w0.png?raw=true
#    :alt: QP results in Si plotted vs the KS energy e0.
#



#!/usr/bin/env python
r"""
Gruneisen parameters with DFPT phonons and finite difference
============================================================

The flow computes the dynamical matrix of AlAs for three different volumens
around the equilibrium configuration.
Gruneisen parameters are then computed in anaddb by approximating the derivative
of the dynamical matrix wrt volume with finite differences.
"""
from __future__ import division, print_function, unicode_literals, absolute_import

import sys
import os
import numpy as np
import abipy.abilab as abilab
import abipy.data as abidata
from abipy import flowtk

def build_flow(options):
    """
    Create a `Flow` for phonon calculations:

        1) One workflow for the GS run.
        2) nqpt works for phonon calculations. Each work contains
           nirred tasks where nirred is the number of irreducible phonon perturbations
           for that particular q-point.
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    flow = flowtk.Flow(workdir=options.workdir)

    # This function constructs the input files for the phonon calculation:
    # GS input + the input files for the phonon calculation.
    pseudos = abidata.pseudos("14si.pspnc")
    structure = abidata.structure_from_ucell("Si")

    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    # List of q-points for the phonon calculation.
    #qpoints = np.reshape(qpoints, (-1, 3))

    # Global variables used both for the GS and the DFPT run.
    gs_inp.set_vars(
        nband=4,
        ecut=2.0,
        ngkpt=[4, 4, 4],
        nshiftk=4,
        shiftk=[0.0, 0.0, 0.5,   # This gives the usual fcc Monkhorst-Pack grid
                0.0, 0.5, 0.0,
                0.5, 0.0, 0.0,
                0.5, 0.5, 0.5],
        #shiftk=[0, 0, 0],
        paral_kgb=0,
        diemac=12.0,
        iomode=3,
        tolvrs=1.0e-18,
    )

    # Get the qpoints in the IBZ. Note that here we use a q-mesh with ngkpt=(4,4,4) and shiftk=(0,0,0)
    qpoints = gs_inp.abiget_ibz(ngkpt=(4, 4, 4), shiftk=(0, 0, 0)).points

    # Build input for GS calculation and register the first work.
    #work0 = flow.register_scf_task(scf_input)
    #ph_work = flowtk.PhononWork.from_scf_task(work0[0], qpoints=qpoints, is_ngqpt=True)
    #flow.register_work(ph_work)

    ph_inputs = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=len(qpoints))

    return flow

# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("GENERATE_SPHINX_GALLERY", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).plot_networkx(tight_layout=True)


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



"""Utilities for pandas dataframe"""
from __future__ import print_function, division, unicode_literals, absolute_import

import sys


def print_dataframe(frame, title=None, precision=6, sortby=None, file=sys.stdout):
    """
    Print entire pandas DataFrame.

    Args:
        frame: pandas DataFrame.
        title: Optional string to print as initial title.
        precision: Floating point output precision (number of significant digits).
            This is only a suggestion [default: 6] [currently: 6]
        sortby: string name or list of names which refer to the axis items to be sorted (dataframe is not changed)
        file: a file-like object (stream); defaults to the current sys.stdout.
    """
    if title is not None: print(title, file=file)
    if sortby is not None and sortby in frame:
        frame = frame.sort_values(sortby, inplace=False)

    import pandas as pd
    with pd.option_context("display.max_rows", len(frame),
                           "display.max_columns", len(list(frame.keys())),
                           "display.precision", precision,
                           ):
        print(frame, file=file)
        print(" ", file=file)
